# -*- coding: utf-8 -*-
"""
Created on Sun Sep  7 13:25:38 2014

@author: Morten
"""
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import *

import numpy as np
from copy import deepcopy
import collections
import logging

from datashape import discover
from datashape.coretypes import (Record, var,Option)
try: 
    from odo import convert 
except ImportError:
    from into import convert
    
from dynd import nd



from ..tools import AnyPyProcessOutputList
logger = logging.getLogger('abt.anypytools')


@convert.register(nd.array, AnyPyProcessOutputList, cost=1.0)
def convert(res, **kwargs):
    # Hack to ensure values are always interpreted as floats
    # This is necessary because the dump command in AnyBody 
    # may cause whole number float to look like int. 
    res = deepcopy(res)
    for elem in res:
        for key in elem:
            if key.startswith('Main') and isinstance(elem[key], np.ndarray):
                try:
                    elem[key] = elem[key].astype('float')    
                except ValueError:
                    pass
    ###
    prepared_data, ds = convert_and_extract_dshape(res, **kwargs)
    return nd.array(prepared_data, dtype=str(ds))


def convert_to_nested_structure(obj):
    assert isinstance(obj, collections.Mapping)
    for key in obj:
        if '.' in key:
            first, last = key.split('.', 1)
            if first not in obj:
                obj[first] = collections.OrderedDict()
            obj[first][last] = obj.pop(key)
            convert_to_nested_structure(obj[first])

def remove_dots_in_key_names(obj):    
    assert isinstance(obj, collections.Mapping)
    for key in list( obj.keys() ):
        obj[key.replace('.','_')] = obj.pop(key)

def convert_items_to_python_types(obj):
    assert isinstance(obj, collections.Mapping)
    for k in obj:
        if isinstance(obj[k], collections.Mapping):
            convert_items_to_python_types(obj[k])
        else:
            if isinstance(obj[k], np.ndarray):
                obj[k] = obj.pop(k).tolist()
            if not obj[k] and isinstance(obj[k], collections.Iterable):
                obj[k] = None

def update_dict_if(d, u, condition = True ):
    """ Update one nested dictionary 'd' with the keys from another 'u'
        If the keys already exist, then values are only if condition
        function returns true
    """       
    if condition is True:
        condition = lambda old_val, new_val: True
    elif condition is False:
        condition = lambda old_val, new_val: False
    for k in u:
        if isinstance(u[k], collections.Mapping):
            r = update_dict_if(d.get(k, {}), u[k], condition)
            d[k] = r
        elif k in d:
            if condition(d[k], u[k]):
                d[k] = u[k]
        else:
            d[k] = u[k]
    return d        


def build_datashape(sample_data):
    """ Build a datashape from sample_data 
        In case of multiple dimensions, the firt is replace by 'var' to 
        handle varing time samples in models
    """
    if isinstance(sample_data, collections.Mapping):
        return Record((k,build_datashape(v)) for k,v in sample_data.items() )
    else:
        ds = discover(sample_data)
        if len(ds) > 1:
            ds = var *  ds.subarray(1)
        else:
            ds = Option( ds )
        return ds

def clean_sample(mapping, key):
    """ Scrubs the values of a nested data structure, leaving the prototype
        empty values. List are replaced by [] and values by None """
    if isinstance(mapping[key], collections.Mapping):
        for subkey in mapping[key]:
            clean_sample(mapping[key], subkey)
    elif isinstance(mapping[key], collections.Iterable):
        mapping[key] = []
    else:
        mapping[key] = None



def convert_and_extract_dshape(result_list, create_nested_structure = False, **kwargs):

    result_list = deepcopy(result_list)
    
    if create_nested_structure:
        for elem in result_list:
            convert_to_nested_structure(elem)
    else:
        for elem in result_list:
            #pass            
            remove_dots_in_key_names(elem)
                        
    for elem in result_list:
        convert_items_to_python_types(elem)



    def update_check(old_value, new_value):
        if isinstance(old_value, collections.Iterable):
            if not any(old_value):
                return True
        elif not old_value:
            return True
        else: 
            return False

    # Create a sample data structure with the keys from all the results.         
    sample_data = collections.OrderedDict()
    for data_structure in result_list:
        update_dict_if(sample_data, data_structure, update_check)
    
    # If some task info is present set this data so the discover function
    # will always find the correct data type. 
    if 'task_id' in sample_data:
        sample_data['task_macro_hash'] = 42
        sample_data['task_id'] = 42
        sample_data['task_work_dir'] = "string"
        sample_data['task_name'] = "string"
        sample_data['task_processtime'] = 10.1
        sample_data['task_macro'] = ['string', 'string']
        sample_data['task_logfile'] = 'string'
        
        
    
    dshape = build_datashape(sample_data)        
        
      
    
    empty_sample_data = deepcopy( sample_data )
    for key in sample_data:
        clean_sample(empty_sample_data, key)
        
        
    for elem in result_list:
        ## Update_dict_if( ) with a always False condition, will
        ## only add missing keys
        update_dict_if(elem, empty_sample_data, condition=False)        
        
        
    
    return (result_list, dshape)  





# -*- coding: utf-8 -*-
"""
    Lexer for the AnyScript modelling langue.

"""
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import *



import os

from pygments.lexer import (RegexLexer, include, bygroups,
                            default, words, inherit)
from pygments.token import (Text, Comment, Operator, Keyword, Name, String,
                            Number, Punctuation, Generic)


__all__ = ['AnyScriptLexer', 'AnyScriptDocLexer']

_ROOT = os.path.abspath(os.path.dirname(__file__))


with open(os.path.join(_ROOT, 'classes.txt')) as f:
    ANYCLASSES = f.read().split('\n')
with open(os.path.join(_ROOT, 'functions.txt')) as f:
    ANYFUNCTIONS = f.read().split('\n')
with open(os.path.join(_ROOT, 'globals.txt')) as f:
    ANYGLOBALS = f.read().split('\n')


class AnyScriptLexer(RegexLexer):
    """
    """
    name = 'AnyScript'
    aliases = ['anyscript']
    filenames = ['*.any']

    # The trailing ?, rather than *, avoids a geometric performance drop here.
    #: only one /* */ style comment
    _ws1 = r'\s*(?:/[*].*?[*]/\s*)?'

    tokens = {
        'whitespace': [
            # # preprocessor directives: without whitespace
            # ('^#if\s+0', Comment.Preproc, 'if0'),
            # ('^#', Comment.Preproc, 'macro'),
            # # or with whitespace
            # ('^(' + _ws1 + r')(#if\s+0)',
             # bygroups(using(this), Comment.Preproc), 'if0'),
            # ('^(' + _ws1 + ')(#)',
             # bygroups(using(this), Comment.Preproc), 'macro'),
            (r'\n', Text),
            (r'\s+', Text),
            (r'\\\n', Text),  # line continuation
            (r'//(\n|(.|\n)*?[^\\]\n)', Comment.Single),
            (r'/(\\\n)?[*](.|\n)*?[*](\\\n)?/', Comment.Multiline),
        ],

        'statements': [
            # For AnyDoc highlighting
            (r'(§)(/[*])(§)((.|\n)*?)(§)([*]/)(§)', 
            bygroups(Generic.Deleted,Generic.Error,Generic.Deleted,Comment.Multiline,Comment.Multiline,Generic.Deleted,Generic.Error,Generic.Deleted)),
            (r'(§)(//)(§)',bygroups(Generic.Deleted,Generic.Error,Generic.Deleted),'multiline-directive'),
            (r'§',Generic.Deleted,'new-codes'),
            ####################
            (words(('#if','#ifdef','#ifndef','#undef','#endif' ,'#include', '#import', '#else', '#elif', '#classtemplate','#define','#path')
             ), Comment.Preproc),
            (r'[L@]?"', String, 'string'),
            (r'(\d+\.\d*|\.\d+|\d+)[eE][+-]?\d+[lL]?', Number.Float),
            (r'(\d+\.\d*|\.\d+|\d+[fF])[fF]?', Number.Float),
            (r'\d+[Ll]?', Number.Integer),
            (r"['&*+=|?:<>/-]", Operator),
            # TODO: "correctly" parse complex code attributes
            (r'[()\[\],.]', Punctuation),
            # Globals
            (words(ANYGLOBALS,
                 suffix=r'\b'), Keyword),
            # Functions
            (words(ANYFUNCTIONS, suffix=r'\b'), Name.Builtin),
            # (r'(\.)([a-zA-Z_]\w*)',
             # bygroups(Operator, Name.Attribute)),
            # void is an actual keyword, others are in glib-2.0.vapi
            (words(ANYCLASSES,
                   suffix=r'\b'), Keyword),
            ('[a-zA-Z_]\w*', Name),
        ],
        'root': [
            include('whitespace'),
            default('statement'),
        ],
        'statement': [
            include('whitespace'),
            include('statements'),
            ('[{}]', Punctuation),
            (';', Punctuation, '#pop'),
        ],
        'string': [
            (r'"', String, '#pop'),
            (r'\\([\\abfnrtv"\']|x[a-fA-F0-9]{2,4}|[0-7]{1,3})', String.Escape),
            (r'[^\\"\n]+', String),  # all other characters
            (r'\\\n', String),  # line continuation
            (r'\\', String),  # stray backslash
        ],
        'multiline-directive': [
            (r'(.*?)(§)', bygroups(Comment.Single,Generic.Deleted), 'new-codes'),
            (r'.*?\n', Comment.Single,'#pop')
        ],
        'new-codes':[
            (r'[^§]+', Generic.Error),
            (r'§', Generic.Deleted, '#pop'),
            (r'[§]', Generic.Error)
        ]        
        # 'macro': [
            # (r'[^/\n]+', Comment.Preproc),
            # (r'/[*](.|\n)*?[*]/', Comment.Multiline),
            # (r'//.*?\n', Comment.Single, '#pop'),
            # (r'/', Comment.Preproc),
            # (r'(?<=\\)\n', Comment.Preproc),
            # (r'\n', Comment.Preproc, '#pop'),
        # ],
        # 'if0': [
            # (r'^\s*#if(?:def).*?(?<!\\)\n', Comment.Preproc, '#push'),
            # (r'^\s*#el(?:se|if).*\n', Comment.Preproc, '#pop'),
            # (r'^\s*#endif.*?(?<!\\)\n', Comment.Preproc, '#pop'),
            # (r'.*?\n', Comment),
        # ]
    }










# -*- coding: utf-8 -*-
"""
Created on Fri Oct 19 21:14:59 2012

@author: Morten
"""
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import *

import logging

from past.builtins import basestring as string_types


import sys
from scipy.stats import distributions
if sys.platform.startswith("win") and sys.version_info.major < 3:
    # This is a horrible hack to work around a bug in
    # scipy http://stackoverflow.com/questions/15457786/ctrl-c-crashes-python-after-importing-scipy-stats
    try:
        import thread #, imp, ctypes, os
    except ImportError:
        import _thread as thread
    import win32api
    def handler(sig, hook=thread.interrupt_main):
        hook()
        return 1
    win32api.SetConsoleCtrlHandler(handler, 1)

import numpy as np
from numpy.random import random, seed

#pprint is used in the doc tests
from .tools import pprint, array2anyscript


from types import GeneratorType
logger = logging.getLogger('abt.anypytools')


def _isgenerator(x):
    return isinstance(x, types.GeneratorType)

class MacroGenerator(object):
    """ Base class for generating AnyScript macros

    Use the class to build an AnyScript macro. The class have methods for all
    anyscript opertions, and makes it easier to construct a macro with the
    correct syntax.

    Examples
    --------
    >>> mg = MacroGenerator()
    >>> mg.add_load('c:/MyModel/model.main.any')
    >>> mg.add_set_value('Main.Study.myvar', 12.3)
    >>> mg.add_run_operation('Main.Study.Kinematics')
    >>> pprint( mg.generate_macros() )
    [['load "c:/MyModel/model.main.any"',
      'classoperation Main.Study.myvar "Set Value" --value="12.3"',
      'operation Main.Study.Kinematics',
      'run']]

    If more than one macro is generated the class can construct a python
    generator object which builds the macro (lazily) as they are requested.


    Parameters
    ----------
    number_of_macros: int
        The number of macros to generate.
    counter_token: str
        A token in the macro commands that will be replace with a counter. If
        the token is '{ID}' then all occurences of '{ID}' in the macros will be
        replaced with a counter.
    reset_counter_for_each_batch: bool
        Specifies if the macro counter is reset if macros are generated in batch
        mode.


    Returns
    -------
    A MacroGenerator object:
        A MacroGenerator object for constructing the macro.
    """
    def __init__(self, number_of_macros=1, counter_token = None,
                 reset_counter_for_each_batch = True):
        assert(number_of_macros > 0)
        self._macro_cmd_list = []
        self.number_of_macros = number_of_macros
        self._counter_token = counter_token
        self._new_batch_resets_counter = reset_counter_for_each_batch
        self.cached_output = None

    def add_macro(self,macro):
        """ Add macro code to the generated macro.

        Parameters
        ----------
        macro:
            A anyscript macro command or a list of anyscript commands


        Examples
        --------
        add a single macro command:

        >>> mg = MacroGenerator()
        >>> mg.add_macro('load "model1.any"')

        add several macro commands:

        >>> mg.add_macro(['operation Main.RunApplication', 'run', 'exit'])


        """
        if not isinstance(macro,list):
            macro = [macro]
        for macro_cmd in macro:
            self._macro_cmd_list.append(macro_cmd)

    def add_set_value(self,variable, value):
        """ Add a 'Set Value' macro command where the value is the same in all
        macros.

        Parameters
        ----------
        variable
            An AnyScript variable or a list of AnyScript variables.
        value
            A Values assign to the AnyScript variable or a list of values.

        Examples
        --------
        Set a single value:

        >>> mg = MacroGenerator()
        >>> mg.add_load('c:/MyModel/model.main.any')
        >>> mg.add_set_value('Main.Study.myvar1', 23.1)
        >>> mg.add_set_value('Main.Study.myvar2', np.array([2,3,4]))
        >>> mg.add_run_operation('Main.Study.Kinematics')
        >>> pprint( mg.generate_macros() )
        [['load "c:/MyModel/model.main.any"',
          'classoperation Main.Study.myvar1 "Set Value" --value="23.1"',
          'classoperation Main.Study.myvar2 "Set Value" --value="{2,3,4}"',
          'operation Main.Study.Kinematics',
          'run']]

        Set variable across different macros

        >>> mg = MacroGenerator(number_of_macros = 3)
        >>> mg.add_load('c:/MyModel/model.main.any')
        >>> mg.add_set_value('Main.Study.myvar1',[1,2,3])
        >>> pprint( mg.generate_macros() )
        [['load "c:/MyModel/model.main.any"', 'classoperation Main.Study.myvar1 "Set Value" --value="1"'],
         ['load "c:/MyModel/model.main.any"', 'classoperation Main.Study.myvar1 "Set Value" --value="2"'],
         ['load "c:/MyModel/model.main.any"', 'classoperation Main.Study.myvar1 "Set Value" --value="3"']]

        The method can also add several macro commands if both variable and
        value are list of the same length

        >>> mg = MacroGenerator()
        >>> mg.add_set_value(['Main.Study.myvar1','Main.Study.myvar1'],[4, 23])
        >>> pprint( mg.generate_macros())
        [['classoperation Main.Study.myvar1 "Set Value" --value="4"',
          'classoperation Main.Study.myvar1 "Set Value" --value="23"']]

        """

        if isinstance(variable,list) and isinstance(value,list):
            if len(variable) != len(value):
                raise ValueError('Lists must be the same length')
            for k,v in zip(variable,value):
                self.add_macro(self._create_set_value_cmd(k,v))

        elif isinstance(variable, string_types) and isinstance(value,list):
            if len(value) != self.number_of_macros:
                raise ValueError("If 'value' is a list it must be the same length\
                                  as the number of macros")
            macro_generator = self._generator_set_value(variable, value)
            self.add_macro(macro_generator)

        else:
            self.add_macro(self._create_set_value_cmd(variable,value))

    def add_set_value_range(self, var, start, stop, endpoint = True ) :
        """ Add a 'Set Value' macro command where the values changes across all
        generated macros. Th values will be evenly spaced over a specified interval.
        Requires that the GenerateMacro object is set to generate more than
        one macro.

        Parameters
        ----------
        start: scalar or array
            The starting value of the sequence.
        stop: scalar or array
            The end value of the sequence, unless endpoint is set to False.
            In that case, the sequence consists of all but the last of num + 1
            evenly spaced samples, so that stop is excluded. Note that the
            step size changes when endpoint is False.
        endpoint: bool, optional
            If True, stop is the last sample. Otherwise, it is not included.
            Default is True.

        Examples:
        ---------
        Set a single value:

        >>> mg = MacroGenerator(number_of_macros = 10)
        >>> mg.add_set_value_range('Main.Study.myvar1', start= 10, stop=100)
        >>> pprint( mg.generate_macros() )
        [['classoperation Main.Study.myvar1 "Set Value" --value="10"'],
         ['classoperation Main.Study.myvar1 "Set Value" --value="20"'],
         ['classoperation Main.Study.myvar1 "Set Value" --value="30"'],
         ['classoperation Main.Study.myvar1 "Set Value" --value="40"'],
         ['classoperation Main.Study.myvar1 "Set Value" --value="50"'],
         ['classoperation Main.Study.myvar1 "Set Value" --value="60"'],
         ['classoperation Main.Study.myvar1 "Set Value" --value="70"'],
         ['classoperation Main.Study.myvar1 "Set Value" --value="80"'],
         ['classoperation Main.Study.myvar1 "Set Value" --value="90"'],
         ['classoperation Main.Study.myvar1 "Set Value" --value="100"']]

        """
        no = self.number_of_macros
        if isinstance(start, np.ndarray):
            assert start.shape == stop.shape, 'Start and stop must be similar'
            valuelist = []
            for start_val, stop_val in zip(start.flatten(), stop.flatten() ):
                arr = np.linspace(start_val,stop_val, self.number_of_macros, endpoint)
                valuelist.append(arr)
            valuelist = np.array(valuelist).T

            valuelist = [subarr.reshape(start.shape) for subarr in valuelist ]
        else:
            valuelist = np.linspace(start,stop, no, endpoint).tolist()

        macro_generator = self._generator_set_value(var, valuelist)
        self.add_macro(macro_generator)


    def _create_set_value_cmd(self,var,val):
        """Creates a set value macro string"""
        if isinstance(val,list):
            val = np.array(val)
        if isinstance(val, np.ndarray):
            val = array2anyscript(val).strip('"')
        if isinstance(val,float):
            val = '{:.12g}'.format(val)
        return 'classoperation {0} "Set Value" --value="{1}"'.format(var,val)

    def _generator_set_value(self, var, values, special_first_values = None):
        """ Creates a generator for the set value macro command """
        if special_first_values is not None:
            yield self._create_set_value_cmd(var, special_first_values)
        for value in values:
            yield self._create_set_value_cmd(var, value)

    def _generator_specific_macro(self, cmd, i_macro):
        """ Generator which only include macro command in the i'th macro"""
        if not isinstance( i_macro, list):
            i_macro = [i_macro]
        for counter in range(self.number_of_macros):
            if counter in i_macro:
                yield cmd
            else:
                yield None


    def add_dump(self, variables, include_in_macro = None):
        """ Create Dump macro command.

        Parameters
        ----------
        variables: string or list of strings
            The anyscript values to create a 'Dump' macro command for
        include_in_macro: integer or list of integers
            Specifices in which macros [0,1,2....NumberOfMacros] to include the
            dump command.
            If None, the command is included in all macros.


        Examples
        --------
        >>> mg = MacroGenerator()
        >>> mg.add_dump('Main.Study.myvar1')
        >>> pprint( mg.generate_macros())
        [['classoperation Main.Study.myvar1 "Dump"']]


        Only include the dump command in the two first macro

        >>> mg = MacroGenerator(number_of_macros = 5)
        >>> mg.add_load('MyModel.any')
        >>> mg.add_dump('Main.Study.myvar1', include_in_macro = [0,1])
        >>> pprint( mg.generate_macros())
        [['load "MyModel.any"', 'classoperation Main.Study.myvar1 "Dump"'],
         ['load "MyModel.any"', 'classoperation Main.Study.myvar1 "Dump"'],
         ['load "MyModel.any"'],
         ['load "MyModel.any"'],
         ['load "MyModel.any"']]

        """
        if not isinstance(variables,list):
            variables = [variables]
        for var in variables:
            if isinstance(var, string_types):
                cmd = 'classoperation {0} "Dump"'.format(var)
                if include_in_macro is not None:
                    self.add_macro( self._generator_specific_macro(cmd, include_in_macro))
                else:
                    self.add_macro(cmd)

    def add_load(self, mainfile, define_kw = {}, path_kw={}):
        """ Create a Load macro command.

        Parameters
        ----------
        mainfile: string
            Path of the main file of the model
        define_kw: dict
            Dictionary of defines statements to set during load
        path_kw: dict
            Dictionary of path staements to set during load
        counter_token: string
            A string in the macro commands which will be replace with a counter
            for every macro generated

        Examples
        --------
        >>> mg = MacroGenerator()
        >>> paths = {'DATA':'c:/MyModel/Data'}
        >>> defines = {'EXCLUDE_ARMS':None, 'N_STEP':20}
        >>> mg.add_load('c:/MyModel/model.main.any', defines, paths)
        >>> pprint( mg.generate_macros())
        [['load "c:/MyModel/model.main.any" -def EXCLUDE_ARMS="" -def N_STEP="20" -p DATA=---"c:/MyModel/Data"']]

        >>> mg = MacroGenerator( number_of_macros = 3, counter_token='{COUNTER}' )
        >>> mg.add_load('c:/MyModel/model_{COUNTER}_.main.any')
        >>> pprint( mg.generate_macros())
        [['load "c:/MyModel/model_0_.main.any"'],
         ['load "c:/MyModel/model_1_.main.any"'],
         ['load "c:/MyModel/model_2_.main.any"']]

        """
        load_cmd = ['load "{}"'.format(mainfile)]

        for key in sorted(define_kw):
            value = define_kw[key]
            if isinstance(value, string_types):
                if value.startswith('"') and value.endswith('"'):
                    value = value[1:-1]
                    load_cmd.append(
                          '-def %s=---"\\"%s\\""'% (key, value.replace('\\','\\\\') ) )
                else:
                    load_cmd.append('-def %s="%s"'% (key, value))
            elif value is None:
                load_cmd.append('-def %s=""'% (key) )
            elif isinstance(value,float) :
                load_cmd.append('-def %s="%g"'% (key, value) )
            else :
                load_cmd.append('-def %s="%d"'% (key, value) )

        for key in sorted( path_kw ):
            load_cmd.append('-p %s=---"%s"'% (key, path_kw[key].replace('\\','\\\\')) )

        self.add_macro(' '.join(load_cmd))


    def add_save_design(self, operation, filename):
        """ Create a Save Design macro command.

        Parameters
        ----------
        operation: string
            The AnyScript operation
        filename: string
            The file in which to save the design

        Examples
        --------
        >>> mg = MacroGenerator()
        >>> mg.add_save_design('Main.MyStudy.Kinematics', 'c:/design.txt')
        >>> pprint( mg.generate_macros())
        [['classoperation Main.MyStudy.Kinematics "Save design" --file="c:/design.txt"']]
        """
        self.add_macro('classoperation {} "Save design" --file="{}"'.format(operation, filename))

    def add_load_design(self, operation, filename):
        """ Create a Load Design macro command.

        Parameters
        ----------
        operation: string
            The AnyScript operation
        filename: string
            The file from which to load the design

        Examples
        --------
        >>> mg = MacroGenerator()
        >>> mg.add_load_design('Main.MyStudy.Kinematics', 'c:/design.txt')
        >>> pprint( mg.generate_macros())
        [['classoperation Main.MyStudy.Kinematics "Load design" --file="c:/design.txt"']]
        """
        self.add_macro('classoperation {} "Load design" '
                       '--file="{}"'.format(operation, filename))

    def add_save_values(self, filename):
        """ Create a Save Values macro command.

        Parameters
        ----------
        filename: string
            The anyset file to save the values to.

        Examples
        --------
        >>> mg = MacroGenerator()
        >>> mg.add_save_values('c:/values.anyset')
        >>> pprint( mg.generate_macros())
        [['classoperation Main "Save Values" --file="c:/values.anyset"']]
        """
        self.add_macro('classoperation Main "Save Values" --file="{}"'.format(filename))

    def add_load_values(self, filename):
        """ Create a Load Values macro command.

        Parameters
        ----------
        filename: string
            The anyset file from which to load the values.

        Examples
        --------
        >>> mg = MacroGenerator()
        >>> mg.add_load_values('c:/values.anyset')
        >>> pprint( mg.generate_macros())
        [['classoperation Main "Load Values" --file="c:/values.anyset"']]
        """
        self.add_macro('classoperation Main "Load Values" --file="{}"'.format(filename))

    def add_update_values(self):
        """ Create a Update Values macro command.

        Examples
        --------
        >>> mg = MacroGenerator()
        >>> mg.add_update_values()
        >>> pprint( mg.generate_macros())
        [['classoperation Main "Update Values"']]
        """
        self.add_macro('classoperation Main "Update Values"')

    def add_run_operation(self, operation):
        """ Create a macro command to select and run an operation

        Examples
        --------
        >>> mg = MacroGenerator()
        >>> mg.add_run_operation('Main.MyStudy.Kinematics')
        >>> pprint( mg.generate_macros())
        [['operation Main.MyStudy.Kinematics', 'run']]
        """
        self.add_macro(['operation {}'.format(operation),'run'])


    def _build_macro(self,i ):
        """ Assemble the macro commands for the i'th macro,  """
        macro = []
        for elem in self._macro_cmd_list:
            if isinstance(elem, string_types):
                macro.append(elem)
            elif isinstance(elem, GeneratorType):
                macro_cmd = next(elem)
                if macro_cmd is not None:
                    macro.append( macro_cmd )
            else:
                continue
            if self._counter_token is not None:
                macro[-1] = macro[-1].replace(self._counter_token, str(i) )
        return macro


    def generate_macros(self, batch_size = None):
        """ Generate the macros. Either as list (batch_size = None) or in batches
        as generator object (memory efficient when generating many macros)

        Examples
        --------
        >>> mg = MacroGenerator(number_of_macros = 4)
        >>> mg.add_load("c:/Model.main.any")
        >>> mg.add_run_operation('Main.study.Kinematics')
        >>> macros =  mg.generate_macros()
        >>> pprint(macros)
        [['load "c:/Model.main.any"', 'operation Main.study.Kinematics', 'run'],
         ['load "c:/Model.main.any"', 'operation Main.study.Kinematics', 'run'],
         ['load "c:/Model.main.any"', 'operation Main.study.Kinematics', 'run'],
         ['load "c:/Model.main.any"', 'operation Main.study.Kinematics', 'run']]


        Generate macros in batches
        >>> mg = MacroGenerator(number_of_macros = 6)
        >>> mg.add_load('c:/Model.main.any')
        >>> macro_gen =  mg.generate_macros(batch_size = 2)
        >>> for i, macros in enumerate( macro_gen ):
        ...     print( 'Batch {}'.format(i) )
        ...     pprint(macros)
        Batch 0
        [['load "c:/Model.main.any"'], ['load "c:/Model.main.any"']]
        Batch 1
        [['load "c:/Model.main.any"'], ['load "c:/Model.main.any"']]
        Batch 2
        [['load "c:/Model.main.any"'], ['load "c:/Model.main.any"']]
        """
        if batch_size is None:
            if not self.cached_output:
                self.cached_output = list(self._macro_generator(0))
            return self.cached_output
        else:
            return self._macro_generator(batch_size)


    def _macro_generator(self, batch_size = 0):
        """ Return a macro generator object"""
        assert(batch_size >= 0)

        macro_batch = []
        for i_macro in range(self.number_of_macros):
            if batch_size == 0:
                yield self._build_macro(i_macro)
            else:
                if self._new_batch_resets_counter:
                    macro_counter = i_macro % batch_size
                else:
                    macro_counter = i_macro
                macro_batch.append( self._build_macro(macro_counter) )
                if macro_counter+1 >= batch_size:
                    yield macro_batch
                    macro_batch = []

        if len(macro_batch):
            yield macro_batch


class MonteCarloMacroGenerator(MacroGenerator):
    """ Generates AnyScript macros for monte carlos studies.

    Class for building AnyScript macros for Monte Carlo parameter studies.
    This class extends the MacroGenerator class with methods, which can randomly
    vary parameters across the generated macros.

    The class contributes the following mehtods:

     - add_set_value_random_uniform()
         Create a set value macro with a uniform random distribution across
         macros

     - add_set_value_random_norm()
         Create a set value macro with a random distribution across
         macros

     - add_set_value_random()
         Create a Set Value operation with custom frozen distribution
         (see  scipy.stats.distributions)

    If 'many' macros should be generated the class can construct a python
    generator object which builds the macro (lazily) as they are requested (see
    the GenerateMacros() method )

    Parameters
    ----------
    number_of_macros:
        The number of macros to generate.

    Returns
    -------
    A MonteCarloMacroGenerator object:
        A MonteCarloMacroGenerator object for constructing the macros


    Examples
    --------
    >>> seed(1)
    >>> mg = MonteCarloMacroGenerator(number_of_macros=4)
    >>> mg.add_load('c:/MyModel/model.main.any')
    >>> mg.add_set_value_random_norm('Main.Study.myvar', means = 2, stdvs = 0.1)
    >>> mg.add_run_operation('Main.Study.Kinematics')
    >>> pprint( mg.generate_macros())
    [['load "c:/MyModel/model.main.any"',
      'classoperation Main.Study.myvar "Set Value" --value="2"',
      'operation Main.Study.Kinematics',
      'run'],
     ['load "c:/MyModel/model.main.any"',
      'classoperation Main.Study.myvar "Set Value" --value="1.97904821591"',
      'operation Main.Study.Kinematics',
      'run'],
     ['load "c:/MyModel/model.main.any"',
      'classoperation Main.Study.myvar "Set Value" --value="2.05838057443"',
      'operation Main.Study.Kinematics',
      'run'],
     ['load "c:/MyModel/model.main.any"',
      'classoperation Main.Study.myvar "Set Value" --value="1.63150523305"',
      'operation Main.Study.Kinematics',
      'run']]

    Generate macros using a custom distribtuion from scipy.stats.distribtuion.
    In this example we use a logistic distribution

    >>> seed(1)
    >>> from scipy.stats.distributions import logistic
    >>> log_dist = logistic( loc= [2,4,10],scale = [0.1,0.5,1] )
    >>> mg = MonteCarloMacroGenerator(number_of_macros=4)
    >>> mg.add_set_value_random('Main.MyVar', log_dist)
    >>> pprint( mg.generate_macros())
    [['classoperation Main.MyVar "Set Value" --value="{2,4,10}"'],
     ['classoperation Main.MyVar "Set Value" --value="{1.96649895478,4.47303588492,0.924084750004}"'],
     ['classoperation Main.MyVar "Set Value" --value="{1.91637851212,3.11986245798,7.71459079162}"'],
     ['classoperation Main.MyVar "Set Value" --value="{1.8525504037,3.68069479813,9.58104766459}"']]

    """

    def __init__(self, *args, **kwargs):
        super(self.__class__,self).__init__(*args, **kwargs)

    def add_set_value_random_uniform(self, variable, means, scale ):
        """ Add a 'Set Value' macro command where the value is chosen from a
        random uniform distribution.

        Parameters
        ----------
        variable: string
            An AnyScript variable or a list of AnyScript variables.
        means: int,float, numpy.ndarray
            The mean value of the random number
        scale: The range of the random variable [ means-scale/2 , means+scale/2]


        Examples
        --------
            Set variable across different macros

        >>> seed(1)
        >>> mg = MonteCarloMacroGenerator(number_of_macros=5)
        >>> mg.add_set_value_random_uniform('Main.Study.myvar', means = 2, scale = 0.1)
        >>> for line in mg.generate_macros(): pprint(line)
        ['classoperation Main.Study.myvar "Set Value" --value="2"']
        ['classoperation Main.Study.myvar "Set Value" --value="1.99170220047"']
        ['classoperation Main.Study.myvar "Set Value" --value="2.02203244934"']
        ['classoperation Main.Study.myvar "Set Value" --value="1.95001143748"']
        ['classoperation Main.Study.myvar "Set Value" --value="1.98023325726"']

        """
        dist = distributions.uniform(means-scale/2.0,scale)
        self.add_set_value_random(variable,dist)

    def add_set_value_random_norm(self, variable, means, stdvs ) :
        """ Add a 'Set Value' macro command where the value is chosen from a
        random normal distribution.

        Parameters
        ----------
        variable: string
            An AnyScript variable or a list of AnyScript variables.
        means: int,float, numpy.ndarray
            The mean value of the random number
        stdvs: The standar deviation of the random variable


        Examples
        --------
            Set variable across different macros

        >>> seed(1)
        >>> mg = MonteCarloMacroGenerator(number_of_macros=5)
        >>> mg.add_set_value_random_norm('Main.Var', means = [1,2,4], stdvs = [0.1,0.5,2])
        >>> for line in mg.generate_macros(): pprint(line)
        ['classoperation Main.Var "Set Value" --value="{1,2,4}"']
        ['classoperation Main.Var "Set Value" --value="{0.979048215908,2.29190287213,-3.36989533908}"']
        ['classoperation Main.Var "Set Value" --value="{0.948229648476,1.47477555917,1.34701845466}"']
        ['classoperation Main.Var "Set Value" --value="{0.910823783045,1.80133318708,3.47655384811}"']
        ['classoperation Main.Var "Set Value" --value="{1.00974531575,1.8980227331,4.96468967866}"']

        """
        dist = distributions.norm(means,stdvs)
        self.add_set_value_random(variable,dist)

    def add_set_value_random(self, variable, frozen_dist ) :
        """ Add a 'Set Value' macro command where the value is chosen from a
        distibution in scipy.stats.distributions.

        Parameters
        ----------
        variable: string
            An AnyScript variable or a list of AnyScript variables.
        frozen_dist: <scipy.stats.distributions.rv_frozen>
            A frozen distribution from scipy.stats.distributions

        Examples
        --------
        >>> seed(1)
        >>> from scipy.stats.distributions import logistic
        >>> log_dist = logistic( loc= [1,3,4],scale = [0.1,0.5,1] )
        >>> mg = MonteCarloMacroGenerator(number_of_macros=4)
        >>> mg.add_set_value_random('Main.MyVar', log_dist)
        >>> for line in mg.generate_macros(): pprint(line)
        ['classoperation Main.MyVar "Set Value" --value="{1,3,4}"']
        ['classoperation Main.MyVar "Set Value" --value="{0.966498954775,3.47303588492,-5.07591525}"']
        ['classoperation Main.MyVar "Set Value" --value="{0.916378512121,2.11986245798,1.71459079162}"']
        ['classoperation Main.MyVar "Set Value" --value="{0.852550403705,2.68069479813,3.58104766459}"']

        """
        if not isinstance(frozen_dist, distributions.rv_frozen):
            raise TypeError('frozen_dist must be frozen distribtuion from \
                            scipy.stats.distributions' )
        mean_value = frozen_dist.mean()
        if isinstance(mean_value,np.ndarray):
            shape = mean_value.shape
        else:
            shape = None
        random_values = (frozen_dist.ppf(random(shape)) for _ in \
                                    range(self.number_of_macros-1) )
        macro_generator = self._generator_set_value(variable,
                                                    random_values,
                                                    special_first_values= mean_value)
        self.add_macro(macro_generator)


class LatinHyperCubeMacroGenerator(MacroGenerator):
    """ Generates AnyScript macros for parameter studies using Latin hyper cube
    sampling  .

    Class for building AnyScript macros for parameter studies with Latin
    Hypercube Sampling (LHS) of the parameter space. The number of generated
    macros determined the number of LHS samples.

    The class uses pyDOE package to generate the LHS data.

    This class extends the MacroGenerator class with the following methods:

     - add_set_value_LHS_uniform()
         Create a Set Value macro  command where the parameters are uniformly
         sampled using Latin Hypercube sampling across the macros

     - add_set_value_LHS()
         Create a Set Value macro command where the parameters are sampled from
         custom distribution using Latin Hypercube sampling across the macros

    If 'many' macros should be generated the class can construct a python
    generator object which builds the macro (lazily) as they are requested (see
    the GenerateMacros() method )

    Parameters
    ----------
    number_of_macros:
        The number of macros to generate.
    criterion: {None, 'c', 'm', 'cm', 'corr'}
        A a string that specifies how points are sampled
        (see: http://pythonhosted.org/pyDOE/randomized.html)
        `None` (Default) points are randomizes within the intervals.
        "center" or "c" center the points within the sampling intervals
        "maximin" or "m": maximize the minimum distance between points, but place
        the point in a randomized location within its interval
        "centermaximin" or "cm": same as “maximin”, but centered within the intervals
        "corr" minimize the maximum correlation coefficient
    iterations: int
        Specifies how many iterations are used to accomplished the selected criterion

    Returns
    -------
    A LatinHyperCubeMacroGenerator object:

    """

    def __init__(self, number_of_macros = 1, criterion = None, iterations = None, **kwargs ):
        super(LatinHyperCubeMacroGenerator, self).__init__(number_of_macros, **kwargs)
        self.LHS_factors = 0
        self.lhd = np.zeros((2,100))
        self.criterion = criterion
        self.iterations = iterations


    def add_set_value_LHS_uniform(self, variable, loc, scale ) :
        """ Add a 'Set Value' macro command where the values are uniformly
        chosen from the  interval [loc - loc + scale]
        using a Latin Hyper Cube sampler.

        Parameters
        ----------
        variable: string
            An AnyScript variable or a list of AnyScript variables.
        loc: int,float, numpy.ndarray
            The start of the interval for uniform sampling.
        scale: The range of the sample interval


        Examples
        --------
        Set variable across different macros

        >>> seed(1)
        >>> mg = LatinHyperCubeMacroGenerator(number_of_macros=8)
        >>> mg.add_set_value_LHS_uniform('Main.myvar1',1,2)
        >>> mg.add_set_value_LHS_uniform('Main.myvar2',10,10)
        >>> pprint( mg.generate_macros() )
        [['classoperation Main.myvar1 "Set Value" --value="2"',
          'classoperation Main.myvar2 "Set Value" --value="15"'],
         ['classoperation Main.myvar1 "Set Value" --value="2.09919186856"',
          'classoperation Main.myvar2 "Set Value" --value="12.6154232435"'],
         ['classoperation Main.myvar1 "Set Value" --value="1.79656505284"',
          'classoperation Main.myvar2 "Set Value" --value="15.6735209175"'],
         ['classoperation Main.myvar1 "Set Value" --value="2.3547986286"',
          'classoperation Main.myvar2 "Set Value" --value="14.1819509088"'],
         ['classoperation Main.myvar1 "Set Value" --value="1.5366889727"',
          'classoperation Main.myvar2 "Set Value" --value="10.9004056168"'],
         ['classoperation Main.myvar1 "Set Value" --value="1.10425550118"',
          'classoperation Main.myvar2 "Set Value" --value="18.5976467955"'],
         ['classoperation Main.myvar1 "Set Value" --value="2.55111306243"',
          'classoperation Main.myvar2 "Set Value" --value="19.5880843877"'],
         ['classoperation Main.myvar1 "Set Value" --value="1.2500285937"',
          'classoperation Main.myvar2 "Set Value" --value="17.1065243755"']]

        """
        if isinstance(loc,list):
            loc = np.array(loc)
        if isinstance(scale,list):
            scale = np.array(scale)
        dist = distributions.uniform(loc,scale)
        self.add_set_value_LHS(variable,dist)


    def add_set_value_LHS(self, var, frozen_dist ) :
        """ Add a 'Set Value' macro command where the values are
        chosen using Latin Hyper Cube Sampling and transformed using custom
        distribution. Note the first generated macro correspond to the mean
        value of the input parameters.

        Parameters
        ----------
        variable: string
            An AnyScript variable or a list of AnyScript variables.
        frozen_dist: <scipy.stats.distributions.rv_frozen>
            A frozen distribution from scipy.stats.distributions


        Examples
        --------
        >>> seed(1)
        >>> from scipy.stats.distributions import norm
        >>> normdist = norm( [1,3,4], [0.1,0.5,1] )
        >>> mg = LatinHyperCubeMacroGenerator(number_of_macros=4)
        >>> mg.add_set_value_LHS('Main.myvar1',normdist)
        >>> mg.add_set_value_LHS('Main.myvar2', normdist)
        >>> pprint( mg.generate_macros() )
        [['classoperation Main.myvar1 "Set Value" --value="{1,3,4}"',
          'classoperation Main.myvar2 "Set Value" --value="{1,3,4}"'],
         ['classoperation Main.myvar1 "Set Value" --value="{1.07895227633,3.41996355574,-0.0241255334168}"',
          'classoperation Main.myvar2 "Set Value" --value="{1.24119096808,2.1047636289,2.00615775617}"'],
         ['classoperation Main.myvar1 "Set Value" --value="{1.01284739969,3.29072197649,5.64666114098}"',
          'classoperation Main.myvar2 "Set Value" --value="{0.970685113414,2.81380147984,4.35758342284}"'],
         ['classoperation Main.myvar1 "Set Value" --value="{0.87423296579,2.54247201378,4.01716347152}"',
          'classoperation Main.myvar2 "Set Value" --value="{1.04333421171,3.13228054445,5.42610272557}"']]

        """

        if not isinstance(frozen_dist, distributions.rv_frozen):
            raise TypeError('frozen_dist must be frozen distribtuion from \
                            scipy.stats.distributions' )

        mean_value = frozen_dist.mean()

        if isinstance(mean_value, np.ndarray):
            shape = mean_value.shape
            n_elem = np.prod(shape)
            lhs_slice = np.s_[self.LHS_factors:self.LHS_factors+n_elem]
            # Create a generator which will later be used to generate the values
            # based on the variables in self.lhd matrix
            values = (frozen_dist.ppf(self.lhd[i , lhs_slice ].reshape(shape) )\
                        for i in  range(self.number_of_macros-1) )
        else:
            #shape = None
            n_elem = 1
            lhs_index = self.LHS_factors
            # Create a generator which will later be used to generate the values
            # based on the variables in self.lhd matrix
            values = (frozen_dist.ppf(self.lhd[i , lhs_index ])\
                        for i in range(self.number_of_macros-1) )

        if self.number_of_macros == 1:
            macro_generator = self._generator_set_value(var,  [mean_value])
        else:
            macro_generator = self._generator_set_value(var,  values,
                                                    special_first_values = mean_value)
        self.add_macro(macro_generator)

        self.LHS_factors += n_elem

    def generate_macros(self, batch_size = None):
        try:
            import pyDOE
        except ImportError:
            raise ImportError('The pyDOE package must be install to use this class')

        # Only generate LHS values if user requested more than macros, since the
        # first macro is allways the mean value
        if self.number_of_macros > 1:
            # Create the Latin hyper cube sample matrix. This is used by the
            # individual macro generator functions when macros are created.
            self.lhd = pyDOE.lhs(self.LHS_factors,  samples=self.number_of_macros-1,
                                 criterion = self.criterion,
                                 iterations = self.iterations)

        return super(LatinHyperCubeMacroGenerator,self).generate_macros(batch_size)



class PertubationMacroGenerator(MacroGenerator):
    """ TODO: Make a class that can generate macros for pertubation studies


    """
    def __init__(self, pertubationfactor = 1e-4):
        super(self.__class__,self).__init__(number_of_macros = 1)

        self.pertubationfactor = pertubationfactor


    def _generator_set_value_designvar(self, var, value, i_designvar):
        for counter in range(self.number_of_macros):
            if counter == i_designvar:
                yield self._create_set_value_cmd(var, value+self.pertubationfactor)
            else:
                yield self._create_set_value_cmd(var, value)


    def add_set_value_designvar(self,var, value):

        macro_generator = self._generator_set_value_designvar(var,value,
                                                              self.number_of_macros)
        self.number_of_macros += 1
        self.add_macro(macro_generator)


    def add_set_value_range(self, var, start, stop, endpoint = True ) :
        no = self.number_of_macros
        if isinstance(start, np.ndarray):
            arr = np.array([np.linspace(i,j, no, endpoint) for i,j in zip(start,stop)])
            arr = arr.T.squeeze()
        else:
            arr = np.linspace(start,stop, no, endpoint)

        values = (_ for _ in arr)

        macro_generator = self._generator_set_value(var, values)
        self.add_macro(macro_generator)


if __name__ == '__main__':

#    mg = PertubationMacroGenerator()
#    mg.add_set_value_designvar('test',1)
#    mg.add_set_value_designvar('test',2)
#
#    pprint(mg.generate_macros())
#
    import pytest
    pytest.main(str('generate_macros.py --doctest-modules'))




import logging
logger = logging.getLogger('abt.anypytools')
logger.addHandler(logging.NullHandler())


from .abcutils import AnyPyProcess
from .generate_macros import (MacroGenerator, MonteCarloMacroGenerator,
                              LatinHyperCubeMacroGenerator)

from .macroutils import AnyMacro
from . import macro_commands

__all__ = ['datautils', 'h5py_wrapper', 'AnyPyProcess',
           'MacroGenerator', 'MonteCarloMacroGenerator', 'LatinHyperCubeMacroGenerator'
           'AnyMacro', 'macro_commands', 'print_versions']

__version__ = '0.9.3'


def print_versions():
    """ Print all the versions of software that AnyPyTools relies on."""
    import sys, platform
    import numpy as np
    print("-=" * 38)
    print("AnyPyTools version: %s" % __version__)
    print("NumPy version: %s" % np.__version__)
    print("Python version: %s" % sys.version)
    (sysname, nodename, release, version, machine, processor) = \
        platform.uname()
    print("Platform: %s-%s-%s (%s)" % (sysname, release, machine, version))
    if sysname == "Linux":
        print("Linux dist: %s" % " ".join(platform.linux_distribution()[:-1]))
    if not processor:
        processor = "not recognized"
    print("Processor: %s" % processor)
    print("Byte-ordering: %s" % sys.byteorder)
    print("-=" * 38)





# -*- coding: utf-8 -*-
"""
Created on Sun Sep  7 13:25:38 2014

@author: Morten
"""
# Python 2/3 compatibility imports
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import *
from past.builtins import basestring as string_types

# Standard lib imports
import os
import re
import sys
import copy
import errno
import pprint as _pprint
import logging
import warnings
import functools
import collections
from ast import literal_eval
from _thread import get_ident as _get_ident

#external imports
import numpy as np


logger = logging.getLogger('abt.anypytools')


# This hacks pprint to always return strings witout u' prefix
# important when running doctest in both python 2 og python 3
class Py3kPrettyPrinter(_pprint.PrettyPrinter):
    def format(self, object, context, maxlevels, level):
        try:
            if isinstance(object, unicode):
                rep = u"'"  + object + u"'"
                return ( rep.encode('utf8'), True, False)
        except NameError:
            pass
        return _pprint.PrettyPrinter.format(self, object, context, maxlevels, level)

def py3k_pprint(s):
    printer = Py3kPrettyPrinter(width = 110)
    printer.pprint(s)

pprint = py3k_pprint



def run_from_ipython():
    try:
        __IPYTHON__
        return True
    except NameError:
        return False


class mixedmethod(object):
    """This decorator mutates a function defined in a class into a 'mixed' class and instance method.

    Usage:
        class Spam:
            @mixedmethod
            def egg(self, cls, *args, **kwargs):
                if self is None:
                    pass # executed if egg was called as a class method (eg. Spam.egg())
                else:
                    pass # executed if egg was called as an instance method (eg. instance.egg())

    The decorated methods need 2 implicit arguments: self and cls, the former being None when
    there is no instance in the call. This follows the same rule as __get__ methods in python's
    descriptor protocol.
    """
    def __init__(self, func):
        self.func = func
    def __get__(self, instance, cls):
        return functools.partial(self.func, instance, cls)



def get_first_key_match(key, names):
    if key in names:
        return key
    matching = [v for v in names if key in v]
    if not matching:
        raise KeyError('The key "{}" could not be found'.format(key))

    if len(matching) > 1:
        print('WARNING: "{}" key is not unique.'
              ' Using the first match'.format(key), file=sys.stderr)
        print('-> ' + matching[0], file=sys.stderr)
        for match in matching[1:]:
            print(' * '+match, file=sys.stderr)

    return matching[0]

class AnyPyProcessOutputList(collections.MutableSequence):
    """ List like class to wrap the output of model simulations.

        The class behaves as a normal list but provide
        extra function to easily access data.
    """
    def __init__(self, *args):
        self.list = list()
        for elem in args:
            self.extend(list(elem))


    def check(self, v):
        if not isinstance(v, collections.MutableSequence) :
            v = [v]
        for e in v:
            if not isinstance(e, collections.OrderedDict):
                raise(TypeError(e))

    def __len__(self): return len(self.list)

    def __getitem__(self, i):
        if isinstance(i, string_types):
            # Find the entries where i matches the keys
            key = get_first_key_match(i, self.list[0])
            try:
                data = np.array(
                    [super(AnyPyProcessOutput, e).__getitem__(key) for e in self.list]
                     )
            except KeyError:
                raise KeyError(" The key '{}' is not present "
                               "in all elements of the output.".format( key ) )
            if data.dtype == np.dtype('O'):
                # Data will be stacked as an array of objects, if the
                # time dimension is not consistant. Warn that some numpy
                # featurs will not be avaiable.
                warnings.warn('\n\nSimulation time varies across macros. '
                      'Numpy does not support ragged arrays. Data is returned  '
                      'as an array of array objects' )
            return data
        else:
            return type(self)(self.list[i]) if isinstance(i, slice) else self.list[i]



    def __delitem__(self, i): del self.list[i]

    def __setitem__(self, i, v):
        self.check(v)
        if isinstance(i, slice):
            self.list[i] = v
        else:
            self.list[i] = v

    def insert(self, i, v):
        self.check(v)
        self.list.insert(i, v)

    def __str__(self):
        return str(self.list)

    def __repr__(self):
        def create_repr(maxlength = 500):
            repr_list = []
            for elem in self.list:
                if not isinstance(elem, AnyPyProcessOutput):
                    repr_list.append( '  ' + _pprint.pformat(elem))
                    continue
                for line in elem._repr_gen(prefix = ' '):
                    repr_list.append(line)
                    if maxlength and len(repr_list) > maxlength:
                        repr_list.append('  ...')
                        return repr_list
                if repr_list and not repr_list[-1].endswith(','):
                    repr_list[-1] = repr_list[-1] + ','

            if len(repr_list):
                repr_list[-1] = repr_list[-1].rstrip(',')
                repr_list[0] = '[' + repr_list[0][1:]
                repr_list[-1] = repr_list[-1] + ']'
            else:
                repr_list.append('[]')
            return repr_list

        repr_str = '\n'.join(create_repr(500))
        if repr_str.endswith('...'):
            np.set_printoptions(threshold = 30)
            repr_str = '\n'.join(create_repr(1000))
            np.set_printoptions()
        return repr_str

    def filter(self, func):
        """ Constructs a AnyPyProcessOutputList object from those elements
        where function returns true.
        """
        return AnyPyProcessOutputList(filter(func, self))


    def to_dynd(self, **kwargs):
        try:
            from .utils.blaze_converter import convert
            return convert(self.list,**kwargs)
        except ImportError:
            raise ImportError('The packages libdynd, dynd-python, datashape, '
                               'odo/into must be installed to convert data ' )

    def shelve(self, filename, key='results'):
        import shelve
        db = shelve.open(filename)
        db[key] = self
        db.close()

    @classmethod
    def from_shelve(cls, filename, key='results'):
        import shelve
        db = shelve.open(filename)
        out = db[key]
        db.close()
        return out



def get_anybodycon_path():
    """  Return the path to default AnyBody console application
    """
    try:
        import winreg
    except ImportError:
        import _winreg as winreg
    try:
        abpath = winreg.QueryValue(winreg.HKEY_CLASSES_ROOT,
                        'AnyBody.AnyScript\shell\open\command')
    except WindowsError:
        raise WindowsError('Could not locate AnyBody in registry')
    abpath = abpath.rsplit(' ',1)[0].strip('"')
    return os.path.join(os.path.dirname(abpath),'AnyBodyCon.exe')


def define2str(key,value=None):
    if isinstance(value, string_types):
        if value.startswith('"') and value.endswith('"'):
            defstr = '-def %s=---"\\"%s\\""'% (key, value[1:-1].replace('\\','\\\\'))
        else:
            defstr = '-def %s="%s"'% (key, value)
    elif value is None:
        defstr = '-def %s=""'% (key)
    elif isinstance(value,float) :
        defstr =  '-def %s="%g"'% (key, value)
    else:
        defstr = '-def %s="%d"'% (key, value)
    return defstr

def path2str(key,path='.'):
    return '-p %s=---"%s"'% (key, path.replace('\\','\\\\'))


def getsubdirs(toppath, search_string = "."):
    """ Find all directories below a given top path.

    Args:
        toppath: top directory when searching for sub directories
        search_string: Limit to directories matching the this regular expression
    Returns:
        List of directories
    """
    if not search_string:
        return [toppath]
    reg_prog = re.compile(search_string)
    dirlist = []
    if search_string == ".":
        dirlist.append(toppath)
    for root, dirs, files in os.walk(toppath):
        for fname in files:
            if reg_prog.search(os.path.join(root,fname)):
                dirlist.append(root)
                continue
    uniqueList = []
    for value in dirlist:
        if value not in uniqueList:
            uniqueList.append(value)
    return uniqueList



def array2anyscript(arr):
    """ Format a numpy array as an anyscript variable
    """
    def tostr(v):
        if np.isreal(v):
            return '{:.12g}'.format(v)
        elif isinstance(v, (string_types, np.str_)):
            return '"{}"'.format(v)

    def createsubarr(arr):
        outstr = ""
        if isinstance(arr, np.ndarray):
            if len(arr) == 1 and not isinstance(arr[0], np.ndarray):
                return '{'+tostr(arr[0]) + '},'
            outstr += '{'
            for row in arr:
                outstr += createsubarr(row)
            outstr = outstr.strip(',') + '},'
            return outstr
        else:
            return outstr + tostr(arr)+','
    if isinstance(arr, np.ndarray) and not arr.shape:
        return tostr(arr.tolist())
    elif isinstance(arr, np.ndarray) :
        return createsubarr(arr).strip(',')
    elif isinstance( arr, float):
        return tostr(arr)
    else:
        return str(arr)

class AnyPyProcessOutput(collections.OrderedDict):
    """Subclassed OrderedDict which supports partial key access"""
    def __getitem__(self,  key ):
        try:
            return super(AnyPyProcessOutput,self).__getitem__(key)
        except KeyError:
            first_key_match = get_first_key_match(key,
                                              super(AnyPyProcessOutput,self).keys())
            return super(AnyPyProcessOutput,self).__getitem__(first_key_match)

    def _repr_gen(self,prefix):
        items = self.items()
        if not items:
            yield prefix + '{}'
            return

        indent = prefix + '{'
        for i, (key,val) in enumerate(items):
            if i == len(self.keys())-1:
                end = '}'
            else:
                end = ','
            key_str = "'"+key+"'" + ': '
            val_str = _pprint.pformat(val)
            if len(prefix) + len(key_str) + len(val_str) < 80:
                yield indent + key_str + val_str + end
            else:
                yield indent + key_str
                indent = prefix + '   '
                for l in val_str.split('\n'):
                   yield indent + l if l.endswith(',') else indent + l + end
            indent = prefix + ' '



    def __repr__(self, _repr_running={}, prefix = '' ):
        call_key = id(self), _get_ident()
        if _repr_running is None:
            _repr_running = {}
        if call_key in _repr_running:
            return '...'
        _repr_running[call_key] = 1
        try:
            if self is None:
                return '%s()' % (self.__class__.__name__,)
            return '\n'.join(self._repr_gen(prefix))
        finally:
            del _repr_running[call_key]


dump_pattern = re.compile(r'Main.*=.*;$')

def parse_anybodycon_output(strvar, errors_to_ignore=None,
                            warnings_to_include = None):
    if errors_to_ignore is None:
        errors_to_ignore = []
    if warnings_to_include is None:
        warnings_to_include = []

    out = AnyPyProcessOutput(  );
    out['ERROR'] = []
    out['WARNING'] = []

    dump_path = None
    for line in strvar.splitlines():
        if '#### Macro command' in line and "Dump" in line:
            me = re.search('Main[^ \"]*', line)
            if me:
                dump_path = me.group(0)
        if dump_pattern.match(line):
            (first, last) = line.split('=',1)
            last = last.strip(' ;')
            var_name = first.strip()
            value_str = last
            if value_str.startswith('{') and value_str.endswith('}'):
                value_str = value_str.replace('{','[').replace('}',']')
            if dump_path:
                var_name = dump_path
                dump_path = None
            try:
                out[var_name.strip()] = literal_eval(value_str)
            except (SyntaxError, ValueError):
                if value_str == '[...]': value_str = '...'
                value_str, nrep = re.subn(r'([^\[\]",\s]+)', r"'''\1'''", value_str)
                if value_str == '': value_str = 'None'
                if value_str.startswith('"') and value_str.endswith('"'):
                    value_str = "'''" + value_str[1:-1] + r"'''"
                try:
                    out[var_name.strip()] = literal_eval(value_str)
                except (SyntaxError, ValueError):
                    out[var_name.strip()] = last
                    warnings.warn('\n\nCould not parse console output:\n'+ line)
        line_has_errors = (line.startswith('ERROR') or line.startswith('Error') or
                           line.startswith('Model loading skipped'))
        if line_has_errors:
            for err_str in errors_to_ignore:
                if err_str in line: break
            else:
                # This is run if we never break,
                #i.e. err was not in the list of errors_to_ignore
                out['ERROR'].append(line)
        line_has_warning = line.startswith(('WARNING','Failed'))
        if line_has_warning:
            for warn_str in warnings_to_include:
                if warn_str in line:
                    out['WARNING'].append(line)
                    break
    # Convert all list object to numpy arrays
    for k,v in out.items():
        if isinstance(v,list):
            out[k] = np.array(v)

    # Move 'ERROR' and 'WARNING' entry to the last position in the ordered dict
    out['WARNING'] = out.pop('WARNING').tolist()
    out['ERROR'] = out.pop('ERROR').tolist()

    # Remove the ERROR/WARNING key if it does not have any entries
    if len(out['ERROR']) == 0:
        del out['ERROR']
    if len(out['WARNING']) == 0:
        del out['WARNING']
    return out



def get_ncpu():
    """ Return the number of CPUs in the computer
    """
    from multiprocessing import cpu_count
    return cpu_count()


def silentremove(filename):
    """ Removes a file ignoring cases where the file does not exits.  """
    try:
        os.remove(filename)
    except OSError as e:
        if e.errno != errno.ENOENT:  # errno.ENOENT : no such file or directory
            logging.debug('Error removing file: ' + filename)
            raise  # re-raise exception if a different error occured



def _run_from_ipython():
    """ Return True if run from IPython
    """
    try:
         __IPYTHON__
         return True
    except NameError:
        return False


def make_hash(o):

  """
  Makes a hash from a dictionary, list, tuple or set to any level, that contains
  only other hashable types (including any lists, tuples, sets, and
  dictionaries).
  http://stackoverflow.com/questions/5884066/hashing-a-python-dictionary
  """

  if isinstance(o, (set, tuple, list)):

    return hash( tuple([make_hash(e) for e in o]) )

  elif not isinstance(o, dict):

    return hash(o)

  new_o = copy.deepcopy(o)
  for k, v in new_o.items():
    new_o[k] = make_hash(v)

  return hash(tuple(frozenset(sorted(new_o.items()))))


_BM_CONSTANTS = {
'ON': '1',
'OFF': '0',
'CONST_MUSCLES_NONE': '0',
'CONST_MUSCLES_SIMPLE': '1',
'CONST_MUSCLES_3E_HILL': '2',
'CONST_HAND_SIMPLE': '0',
'CONST_HAND_DETAILED': '1',
'CONST_LEG_MODEL_OFF': '"OFF"',
'CONST_LEG_MODEL_Leg': '"Leg"',
'CONST_LEG_MODEL_TLEM': '"TLEM"',
'CONST_MORPH_NONE': '0',
'CONST_MORPH_TRUNK_TO_LEG': '1',
'CONST_MORPH_LEG_TO_TRUNK': '2',
'CONST_PELVIS_DISPLAY_NONE': '0',
'CONST_PELVIS_DISPLAY_LEGPELVIS_ONLY': '1',
'CONST_PELVIS_DISPLAY_LEGANDTRUNKPELVIS': '2',
'CONST_SCALING_CUSTOM': '-1',
'CONST_SCALING_STANDARD': '0',
'CONST_SCALING_UNIFORM': '1',
'CONST_SCALING_LENGTHMASS': '2',
'CONST_SCALING_LENGTHMASSFAT': '3',
'CONST_SCALING_UNIFORM_EXTMEASUREMENTS': '4',
'CONST_SCALING_LENGTHMASS_EXTMEASUREMENTS': '5',
'CONST_SCALING_LENGTHMASSFAT_EXTMEASUREMENTS': '6',
'CONST_SCALING_LENGTHMASSFAT_MULTIDOFS': '7',
}

def replace_bm_constants(d):
    for k, v in d.items():
        if v in _BM_CONSTANTS:
            d[k] = _BM_CONSTANTS[v]
    return d



# -*- coding: utf-8 -*-
"""
Created on Mon Jan 16 11:40:42 2012

@author: mel
"""
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import *

__all__ = ['anydatah5_generator','anyoutputfile_generator',
           'read_anyoutputfile']
import os
import logging
import os.path as op

import numpy as np
#from scipy.interpolate import interp1d
from ast import literal_eval
import warnings
import re

logger = logging.getLogger('abt.anypytools')



def anydatah5_generator(folder=None, match = '.*(\.h5)'):
    """ Create a generator which opens anydata.h5 files

    Parameters
    ----------
    folder : str, optional
        path of the directory to look for anydata.h5 files. Defaults is the
        current working directory
    match : str
        regular expression to match with filenames. Default is to include
        file ending with .h5 '.*(\.h5)'

    Yields
    -------
    h5py_file : array_like
        a h5_py file opened with anypytools.h5py_wrapper
    filename : str
        name of the file from which the data was read


    Examples
    --------
    >>> for (h5, filename) in anydatah5_generator():
            print(h5)

    """
    from . import h5py_wrapper
    if folder is None:
        folder = str( os.getcwd() )
    def func(item):
        return re.match(item, match, flags=re.IGNORECASE)
    filelist = filter(func,  os.listdir(folder))
    for filename in filelist:
        try:
            with h5py_wrapper.File(op.join(folder,filename)) as h5file:
                yield (h5file, filename)
        except IOError:
            pass

def anyoutputfile_generator(folder = None, match = '.*\.(csv|txt)'):
    """ Create a generator which opens AnyOutput files

    Parameters
    ----------
    folder : str, optional
        path of the directory to look for AnyOutput files. Defaults is the
        current working directory
    match : str
        regular expression to match with filenames. Default is to include
        file ending with .csv or .txt '.*(.csv|.txt)'

    Yields
    -------
    data : array_like
        data collumns  in the anyoutput file
    header : list
        list of the collumn names
    constants : dict
        mapping with constants found in the file
    filename : str
        name of the file from which the data was read


    Examples
    --------
    >>> for (data, header, const, fn) in anyoutputfile_generator():
            print(fn)
            print(header)
            print(data.shape)
    """

    if folder is None:
        folder = str(os.getcwd())

    def func(item):
        return re.match(item, match, flags=re.IGNORECASE)
    filelist = filter(func, os.listdir(folder))

    for filename in filelist:
        filepath = op.join(folder,filename)
        data, header, const = open_anyoutputfile(filepath)
        if data is None:
            continue

        yield (data, header, const, os.path.basename(filepath))

def open_anyoutputfile(filepath):
    warnings.warn("Deprecated: open_anyoutputfile is renamed to "
                  "read_anyoutputfile", DeprecationWarning)
    return read_anyoutputfile(filepath)


def read_anyoutputfile(filepath):
    """ Read an AnyOutput file

    Parameters
    ---------
    filepath : str


    Returns
    -------
    data : array_like
        data collumns  in the anyoutput file
    header : list
        list of the collumn names
    constants : dict
        mapping with constants found in the file


    """

    def is_scinum(str):
        try:
            np.float(str)
            return True
        except ValueError:
            return False

    with open(filepath,'rU') as anyoutputfile:
        constants = {}
        reader = iter(anyoutputfile.readline, b'')
        #Check when the header section ends
        fpos1 = 0
        fpos0 = 0
        for row in reader:
            if is_scinum(row.split(',')[0]):
                break
            #Save constant from AnyOutput file
            const,value= _parse_anyoutputfile_constants(row)
            if const is not None:
                constants[const] = value
            fpos1, fpos0 = fpos0, anyoutputfile.tell()
        else:
            warnings.warn("No numeric data in " + os.path.basename(filepath))
            return (None,None,constants)
        # Read last line of the header section if there is a header
        if fpos0 != 0:
            anyoutputfile.seek(fpos1)
            header = next(reader).strip('\n').split(',')
        else:
            header = None

        data = []
        for row in reader:
            try:
                data.append([float(val) for val in row.strip('\n').split(',')])
            except ValueError:
                break
        data = np.array(data)
    return (data, header, constants)



def _parse_anyoutputfile_constants(strvar):
    value = None
    varname = None
    if strvar.count('=') == 1 and strvar.startswith('Main'):
        (first, last) = strvar.split('=')
        varname = first.strip()
        last = last.strip()
        value = None
        last = last.strip('\n')
        if last.find('{') == -1:
            try:
                value = str(literal_eval("'''"+last+"'''") )
                value = str(literal_eval(last))
                value = float(literal_eval(last))
            except:
                pass
        else:
            last = last.replace('{','[').replace('}',']')
            try:
                value = np.array(literal_eval("'''"+last+"'''") )
                value = np.array(literal_eval(last))
            except:
                pass
    return (varname, value)



#def interp_percent(data, indices):
#    data = np.array(data)
#    data = data.squeeze()
#    indices = np.array(indices)
#    x = np.linspace(0,100,len(indices))
#    xnew = np.arange(0,100)
#    y = data[indices]
#    ipolfun = interp1d(x,y)
#    return ipolfun(xnew)

#def any_eval(string):
#
#    string = string.split('//',1)[0]
#    string  = string.split("=",1)[-1].split(";",1)[0].strip()
#
#    string = string.replace('{','[')
#    string = string.replace('}',']')
#    return literal_eval(string)



if __name__ == '__main__':
#    for data,header,filename in csv_trial_data('C:\Users\mel\SMIModelOutput', DEBUG= True):
#        print header

    outvars = ['/Output/Validation/EMG']




# -*- coding: utf-8 -*-
"""
    Pygment style for the AnyScript modelling langue.

"""
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import *


from pygments.style import Style
from pygments.token import Keyword, Name, Comment, String, Error, Number, Operator, Generic, Whitespace, Token 

__all__ = ['AnyScriptStyle']

class AnyScriptStyle(Style):
    background_color="#f8f8f8"
    default_style = ""

    styles={
          Whitespace:                "#bbbbbb", 
          Comment:                   "noitalic #4AA02C", 
          Comment.Preproc:           "noitalic #0000FF", 
   
          #Keyword:                   "bold #AA22FF", 
          Keyword:                   "#0000FF", 
          Keyword.Pseudo:            "nobold", 
          Keyword.Type:              "nobold #0000FF", 
   
          Operator:                  "#111111", 
          Operator.Word:             "bold #AA22FF", 
   
          Name.Builtin:              "#0000FF", 
          Name.Function:             "#0000FF", 
          Name.Class:                "#0000FF", 
          Name.Namespace:            "#0000FF", 
          Name.Exception:            "#D2413A", 
          Name.Variable:             "#19177C", 
          Name.Constant:             "#880000", 
          Name.Label:                "#A0A000", 
          Name.Entity:               "bold #999999", 
          Name.Attribute:            "#7D9029", 
          Name.Tag:                  "bold #008000", 
          Name.Decorator:            "#AA22FF", 
   
          String:                    "#666666", 
          String.Doc:                "italic", 
          String.Interpol:           "bold #BB6688", 
          String.Escape:             "bold #BB6622", 
          String.Regex:              "#BB6688", 
          #String.Symbol:             "#B8860B", 
          String.Symbol:             "#19177C", 
          String.Other:              "#008000", 
          Number:                    "#666666", 
   
          Generic.Heading:           "bold #000080", 
          Generic.Subheading:        "bold #800080", 
          Generic.Deleted:           "#f8f8f8", #Used for $ tag in AnyScript.
          Generic.Inserted:          "#00A000", 
          Generic.Error:             "#FF0000", 
          Generic.Emph:              "italic", 
          Generic.Strong:            "bold", 
          Generic.Prompt:            "bold #000080", 
          Generic.Output:            "#888", 
          Generic.Traceback:         "#04D", 
   
          Error:                     "border:#FF0000" 


    }


# -*- coding: utf-8 -*-
"""
Utilities for working with the AnyBody Console applicaiton

Created on Fri Oct 19 21:14:59 2012
@author: Morten
"""
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import *  # nopep8
import os
import sys
import time
import copy
import h5py
import types
import ctypes
import shelve
import atexit
import logging
import warnings
import collections
import numpy as np
from subprocess import Popen
from tempfile import NamedTemporaryFile
from threading import Thread, RLock
from queue import Queue

from future.utils import text_to_native_str
from past.builtins import basestring as string_types

from .tools import (make_hash, AnyPyProcessOutputList, parse_anybodycon_output,
                    getsubdirs, get_anybodycon_path, mixedmethod,
                    AnyPyProcessOutput, run_from_ipython, get_ncpu, silentremove)
from .macroutils import AnyMacro, MacroCommand

try:
    from IPython.display import HTML, display 
    import ipywidgets
except ImportError:
    ipywidgets = HTML = display = None


logger = logging.getLogger('abt.anypytools')

_thread_lock = RLock()
_KILLED_BY_ANYPYTOOLS = -10


class _SubProcessContainer(object):
    """ Class to hold a record of process pids from Popen.

        Properties
        ----------
        stop_all: boolean
            If set to True all process held by the object will be automatically
            killed

        Methods
        -------
        add(pid):
            Add process id to the record of process

        remove(pid):
            Remove process id from the record

    """

    def __init__(self):
        self._pids = set()
        self._stop_all = False

    def add(self, pid):
        with _thread_lock:
            self._pids.add(pid)
        if self.stop_all:
            self._kill_running_processes()

    def remove(self, pid):
        with _thread_lock:
            try:
                self._pids.remove(pid)
            except KeyError:
                pass

    @property
    def stop_all(self):
        return self._stop_all

    @stop_all.setter
    def stop_all(self, value):
        with _thread_lock:
            if value:
                self._stop_all = True
                self._kill_running_processes()
            else:
                self._stop_all = False

    def _kill_running_processes(self):
        """ Clean up and shut down any running processes
        """
        # Kill any rouge processes that are still running.
        with _thread_lock:
            killed = []
            for pid in self._pids:
                try:
                    os.kill(pid, _KILLED_BY_ANYPYTOOLS)
                    killed.append(str(pid))
                except:
                    pass
            self._pids.clear()


_subprocess_container = _SubProcessContainer()
atexit.register(_subprocess_container._kill_running_processes)


def _display(line, *args, **kwargs):
    if run_from_ipython():
        display(HTML(line))
    else:
        print(line, *args, **kwargs)


def _execute_anybodycon(macro,
                        logfile,
                        anybodycon_path=None,
                        timeout=3600,
                        keep_macrofile=False,
                        env=None):
    """ Launches the AnyBodyConsole applicaiton with the specified macro
        saving the result to logfile """
    if anybodycon_path is None:
        anybodycon_path = get_anybodycon_path()
    if not os.path.isfile(anybodycon_path):
        raise IOError("Can not find anybodycon.exe: " + anybodycon_path)
    macro_filename = os.path.splitext(logfile.name)[0] + '.anymcr'
    with open(macro_filename, 'w+b') as macro_file:
        macro_file.write('\n'.join(macro).encode('UTF-8'))
        macro_file.flush()
    anybodycmd = [os.path.realpath(anybodycon_path),
                  '--macro=', macro_file.name, '/ni']
    if sys.platform.startswith("win"):
        # Don't display the Windows GPF dialog if the invoked program dies.
        # See comp.os.ms-windows.programmer.win32
        # How to suppress crash notification dialog?, Jan 14,2004 -
        # Raymond Chen's response [1]
        SEM_NOGPFAULTERRORBOX = 0x0002  # From MSDN
        ctypes.windll.kernel32.SetErrorMode(SEM_NOGPFAULTERRORBOX)
        subprocess_flags = 0x8000000  # win32con.CREATE_NO_WINDOW?
    else:
        subprocess_flags = 0
    try:
        # Check global module flag to avoid starting processes after
        # the user cancelled the processes
        proc = Popen(anybodycmd,
                     stdout=logfile,
                     stderr=logfile,
                     creationflags=subprocess_flags,
                     env=env)
        _subprocess_container.add(proc.pid)
        timeout_time = time.clock() + timeout
        while proc.poll() is None:
            if time.clock() > timeout_time:
                proc.terminate()
                proc.communicate()
                logfile.seek(0, 2)
                logfile.write('ERROR: Timeout after {:d} sec.'.format(timeout))
                break
            time.sleep(0.05)
        _subprocess_container.remove(proc.pid)
        if proc.returncode == _KILLED_BY_ANYPYTOOLS:
            logfile.write('Anybodycon.exe was interrupted by AnyPyTools')
        elif proc.returncode:
            logfile.write('ERROR: anybodycon.exe exited unexpectedly.'
                          ' Return code: ' + str(proc.returncode))
        if not keep_macrofile:
            silentremove(macro_file.name)
    finally:
        logfile.seek(0)
    return proc.returncode


class _Task(object):
    """Class for storing processing jobs

    Attributes:
        folder: directory in which the macro is executed
        macro: list of macro commands to execute
        number: id number of the task
        name: name of the task, which is used for printing status informations
    """

    def __init__(self, folder=None, macro=None,
                 taskname=None, number=1):
        """ Init the Task class with the class attributes
        """
        self.folder = folder
        if not folder:
            self.folder = os.getcwd()
        if macro is not None:
            self.macro = macro
        else:
            self.macro = []
        self.output = AnyPyProcessOutput()
        self.number = number
        self.logfile = ""
        self.processtime = 0
        self.name = taskname
        if not taskname:
            head, folder = os.path.split(folder)
            parentfolder = os.path.basename(head)
            self.name = parentfolder + '/' + folder

    @property
    def has_error(self):
        return 'ERROR' in self.output

    def add_error(self, error_msg):
        try:
            self.output['ERROR'].append(error_msg)
        except KeyError:
            self.output['ERROR'] = [error_msg]

    def get_output(self, include_task_info=True):
        out = self.output
        if include_task_info:
            out['task_macro_hash'] = make_hash(self.macro)
            out['task_id'] = self.number
            out['task_work_dir'] = self.folder
            out['task_name'] = self.name
            out['task_processtime'] = self.processtime
            out['task_macro'] = self.macro
            out['task_logfile'] = self.logfile
        return out

    @classmethod
    def from_output_data(cls, task_output):
        if not cls.is_valid(task_output):
            raise ValueError('Output can only be reprocessed, if "Task info"'
                             'is included in the output.')
        task = cls(folder=task_output['task_work_dir'],
                   macro=task_output['task_macro'],
                   taskname=task_output['task_name'],
                   number=task_output['task_id'])
        task.processtime = task_output['task_processtime']
        task.output = task_output
        return task

    @classmethod
    def from_output_list(cls, outputlist):
        for elem in outputlist:
            yield cls.from_output_data(elem)

    @classmethod
    def from_macrofolderlist(cls, macrolist, folderlist):
        if not macrolist:
            raise StopIteration
        macrofolderlist = ((m, f) for f in folderlist for m in macrolist)
        for i, (macro, folder) in enumerate(macrofolderlist):
            yield cls(folder, macro, number=i)

    @staticmethod
    def is_valid(output_elem):
        keys = ('task_macro_hash', 'task_id', 'task_work_dir', 'task_name',
                'task_processtime', 'task_macro', 'task_logfile')
        return all(k in output_elem for k in keys)


class _Summery(object):
    """ class to display the summery of task """

    def __init__(self, have_ipython=False, silent=False):
        self._silent = silent
        if have_ipython and ipywidgets and not self._silent:
            self.ipywidget = ipywidgets.HTML()
            self.ipywidget.initialized = False
        else:
            self.ipywidget = None

    def task_summery(self, task):
        if self.ipywidget:
            if task.has_error:
                self._display(self.format_summery(task))

    def _display(self, s):
        if self._silent:
            return
        if self.ipywidget is not None:
            if not self.ipywidget.initialized:
                display(self.ipywidget)
                self.ipywidget.initialized = True
            self.ipywidget.value += s + '<br>'
        else:
            print(s)

    def format_summery(self, task):
        entry = ''
        if task.has_error:
            entry += 'Failed :'
        elif task.processtime == 0:
            entry += 'Not completed :'
        else:
            entry += 'Completed :'
        entry += '{1!s} : {2:5.0f} sec : {0} : '.format(task.name,
                                                        task.number,
                                                        task.processtime)
        if task.logfile:
            if run_from_ipython():
                tmpl = '<a href="file:///{0}" target="_blank">{1}</a>'
                entry += tmpl.format(task.logfile,
                                     os.path.basename(task.logfile))
            else:
                entry += '{0}'.format(os.path.basename(task.logfile))
        return entry

    def final_summery(self, total_process_time, tasklist):
        unfinished_tasks = [t for t in tasklist if t.processtime <= 0]
        failed_tasks = [t for t in tasklist
                        if t.has_error and t.processtime > 0]
        if len(failed_tasks):
            self._display('Tasks with errors: {:d}'.format(len(failed_tasks)))
            if self.ipywidget is None:
                self._display('\n'.join([self.format_summery(t)
                                         for t in failed_tasks]))
        if len(unfinished_tasks):
            self._display('Tasks that did not complete: '
                          '{:d}'.format(len(unfinished_tasks)))
        self._display('Total time: {:.1f} seconds'.format(total_process_time))


class AnyPyProcess(object):
    """
    Class for configuring batch process jobs of AnyBody models.

    This is the main interface to control the AnyBody console application from
    python. The class stores all the configuration about how AnyBody is run.
    It has one important method `start_macro` which launches the AnyBody with
    a given anyscript macro.

    Parameters
    ----------
    num_processes : int, optional
        Number of anybody models to start in parallel.
        This defaults to the number of logical CPU cores in the computer.
    anybodycon_path : str, optional
        Overwrite the default anybodycon.exe file to
        use in batch processing. Defaults to what is found in the windows
        registry.
    timeout : int, optional
        Maximum time (i seconds) a model can run until it is terminated.
        Defaults to 1 hour (3600 sec)-
    ignore_errors : list of str, optional
        List of AnyBody Errors substrings to ignore when running the models.
    return_task_info : bool, optional
        Return the task status information when running macros. Defaults to False
    disp
        Set to False to suppress output (deprecated)
    silent : bool, optional
        Set to True to suppres any output (progress bar and error messages).
    warnings_to_include : list of str, optional
        List of strings that are matched to warnings in the model
        output. If a warning with that string is found the warning
        is returned in the output.

    Returns
    -------
    app instance : AnyPyProcess :
        An instance of the AnyPyProcess object for running batch processing,
        parameter studies and pertubation jobs.

    Examples
    --------
    >>> app = AnyPyProcess(num_processes=8, return_task_info=True)

    """

    def __init__(self,
                 num_processes=get_ncpu(),
                 anybodycon_path=None,
                 timeout=3600,
                 silent=False,
                 ignore_errors=None,
                 warnings_to_include=None,
                 return_task_info=False,
                 keep_logfiles=False,
                 logfile_prefix=None,
                 python_env=None,
                 **kwargs):

        if not isinstance(ignore_errors, (list, type(None))):
            raise ValueError('ignore_errors must be a list of strings')

        if not isinstance(warnings_to_include, (list, type(None))):
            raise ValueError('warnings_to_include must be a list of strings')

        if anybodycon_path is None:
            self.anybodycon_path = get_anybodycon_path()
        elif os.path.exists(anybodycon_path):
            self.anybodycon_path = anybodycon_path
        else:
            raise FileNotFoundError("Can't find " + anybodycon_path)
        self.num_processes = num_processes
        self.silent = silent
        if 'disp' in kwargs:
            warnings.warn("Using 'disp' is deprecated. Use "
                          "AnyPyProcess(silent=True) instead.",
                          DeprecationWarning)
            if kwargs.pop('disp') is False:
                self.silent = True
            else:
                self.silent = False
        self.timeout = timeout
        self.counter = 0
        self.return_task_info = return_task_info
        self.ignore_errors = ignore_errors
        self.warnings_to_include = warnings_to_include
        self.keep_logfiles = keep_logfiles
        if logfile_prefix is not None:
            self.logfile_prefix = logfile_prefix + '_'
        else:
            self.logfile_prefix = logfile_prefix
        self.cached_arg_hash = None
        self.cached_tasklist = None
        if python_env is not None:
            if not os.path.isdir(python_env):
                raise FileNotFoundError('Python environment does'
                                        ' not exist:' + python_env)
            env = dict(os.environ)
            env['PYTHONHOME'] = python_env
            env['PATH'] = env['PYTHONHOME'] + ';' + env['PATH']
            self.env = env
        else:
            self.env = None
        logging.debug('\nAnyPyProcess initialized')

    def save_results(self, filename, append=False):
        if self.cached_tasklist:
            savekey = text_to_native_str('processed_tasks')
            db = shelve.open(text_to_native_str(filename), writeback=True)
            if not append or savekey not in db:
                db[savekey] = self.cached_tasklist
            else:
                db[savekey].extend(self.cached_tasklist)
            db.close()
        else:
            raise ValueError('Noting to save')

    def save_to_hdf5(self, filename, batch_name):
        if self.cached_tasklist:
            any_output = [task.get_output() for task in self.cached_tasklist]
            any_output = AnyPyProcessOutputList(any_output)
            with h5py.File(filename, "w") as f:  # , compression="gzip" # Not sure how much gzip effects reading speed.
                group = f.create_group(batch_name)
                task_names = []
                for run in any_output:
                    task_names.append(run['task_name'])
                # If task names are unique, these will be used as run folder names
                if len(task_names) == len(set(task_names)):
                    for run in any_output:
                            if len(task_names) == len(set(task_names)):
                                task_name = run['task_name'].replace('/', '|')
                                new_folder = group.create_group(task_name)
                                for k, v in run.items():
                                    if not isinstance(v, np.ndarray):
                                        if isinstance(v, list):
                                            new_folder.attrs[str(k)] = str(v)
                                        else:
                                            new_folder.attrs[str(k)] = v
                                    elif isinstance(v, np.ndarray):
                                        new_folder.create_dataset(str(k), data=v)
                # If task names are not unique, task id's will be used as run folder names
                elif len(task_names) != len(set(task_names)):
                    for run in any_output:
                        task_id = str(run['task_id'])
                        new_folder = group.create_group(task_id)
                        for k, v in run.items():
                            if not isinstance(v, np.ndarray):
                                    if isinstance(v, list):
                                        new_folder.attrs[str(k)] = str(v)
                                    else:
                                        new_folder.attrs[str(k)] = v
                            elif isinstance(v, np.ndarray):
                                    new_folder.create_dataset(str(k), data=v)

    @mixedmethod
    def load_results(self, cls, filename):
        loadkey = text_to_native_str('processed_tasks')
        db = shelve.open(text_to_native_str(filename))
        loaded_data = db[loadkey]
        db.close()
        # Hack to help Enrico convert data to the new structured
        if not isinstance(loaded_data[0].output, AnyPyProcessOutput):
            for task in loaded_data:
                task.output = AnyPyProcessOutput(task.output)
        # Check if the functions is called as an instance method.
        if self is not None:
            self.cached_tasklist = loaded_data
        results = [task.get_output(True) for task in loaded_data]
        return AnyPyProcessOutputList(results)

    def start_macro(self, macrolist=None, folderlist=None, search_subdirs=None,
                    **kwargs):
        """Starts a batch processing job.

        Runs a list of AnyBody Macro commands in
        the current directory, or in the folders specified by `folderlist`. If
        `search_subdirs` is a regular expression the folderlist will be expanded
        to include all subdirectories that match the regular expression

        Parameters
        ----------
        macrolist : list of macrocommands, optional
            List of anyscript macro commands. This may also be obmitted in
            which case the previous macros will be re-run.
        folderlist : list of str, optional
            List of folders in which to excute the macro commands. If `None` the
            current working directory is used. This may also be a list of
            tuples to specify a name to appear in the output
        search_subdirs : str, optional
            Regular expression used to extend the folderlist with all the
            subdirectories that match the regular expression.
            Defaults to None: No subdirectories are included.

        Returns
        -------
        list
            A list with the output from each macro executed. This maybe empty
            the macros did not output any data.


        Examples
        --------

        >>> macro = [['load "model1.any"', 'operation Main.RunApplication', 'run'],
                     ['load "model2.any"', 'operation Main.RunApplication', 'run'],
                     ['load "model3.any"', 'operation Main.RunApplication', 'run']]
        >>> folderlist = [('path1/', 'name1'), ('path2/', 'name2')]
        >>> start_macro(macrolist, folderlist, search_subdirs = "*.main.any")

        """
        # Handle different input types
        if isinstance(macrolist, types.GeneratorType):
            macrolist = list(macrolist)
        if isinstance(macrolist, AnyMacro):
            macrolist = macrolist.create_macros()
        elif isinstance(macrolist, list) and len(macrolist):
            if isinstance(macrolist[0], string_types):
                macrolist = [macrolist]
            if isinstance(macrolist[0], MacroCommand):
                macrolist = [macrolist]
            if isinstance(macrolist[0], list) and len(macrolist[0]):
                if isinstance(macrolist[0][0], MacroCommand):
                    macrolist = [[mc.get_macro(index=0) for mc in elem]
                                 for elem in macrolist]
        elif isinstance(macrolist, string_types):
            if macrolist.startswith('[') and macrolist.endswith(']'):
                macrolist = macrolist.strip('[').rstrip(']')
                macrolist = [macrolist.split(', ')]
            else:
                macrolist = [[macrolist]]
        elif isinstance(macrolist, (type(None), AnyPyProcessOutputList)):
            pass
        else:
            raise ValueError('Wrong input argument for macrolist')
        # Check folderlist input argument
        if not folderlist:
            folderlist = [os.getcwd()]
        if not isinstance(folderlist, list):
            raise TypeError('folderlist must be a list of folders')
        # Extend the folderlist if search_subdir is given
        if (isinstance(search_subdirs, string_types) and
                isinstance(folderlist[0], string_types)):
            folderlist = sum([getsubdirs(d, search_subdirs)
                              for d in folderlist], [])
        # Check the input arguments and generate the tasklist
        if macrolist is None:
            if self.cached_tasklist:
                tasklist = self.cached_tasklist
            else:
                raise ValueError('macrolist argument can only be ommitted if '
                                 'the AnyPyProcess object has cached output '
                                 'to process')
        elif isinstance(macrolist[0], collections.Mapping):
            tasklist = list(_Task.from_output_list(macrolist))
        elif isinstance(macrolist[0], list):
            arg_hash = make_hash([macrolist, folderlist, search_subdirs])
            if self.cached_tasklist and self.cached_arg_hash == arg_hash:
                tasklist = self.cached_tasklist
            else:
                self.cached_arg_hash = arg_hash
                tasklist = list(_Task.from_macrofolderlist(macrolist,
                                                           folderlist))
        else:
            raise ValueError('Nothing to process for ' + str(macrolist))

        self.summery = _Summery(have_ipython=run_from_ipython(),
                                silent=self.silent)

        if self.logfile_prefix is None:
            self.logfile_prefix = str(self.cached_arg_hash)[:4] + '_'

        # Start the scheduler
        process_time = self._schedule_processes(tasklist, self._worker)
        self.cleanup_logfiles(tasklist)
        # Cache the processed tasklist for restarting later
        self.cached_tasklist = tasklist
        self.summery.final_summery(process_time, tasklist)
        task_output = [task.get_output(include_task_info=self.return_task_info)
                       for task in tasklist]
        return AnyPyProcessOutputList(task_output)

    #    def _print_summery(self, tasks, duration):
    #        unfinished_tasks = [t for t in tasks if t.processtime <= 0]
    #        failed_tasks = [t for t in tasks if t.has_error and t.processtime > 0]
    #        if len(failed_tasks):
    #            _display('Tasks with errors: {:d}'.format(len(failed_tasks)))
    #            if not run_from_ipython():
    #                _display('\n'.join([t.summery() for t in failed_tasks]))
    #        if len(unfinished_tasks):
    #            _display('Tasks that did not complete: '
    #                     '{:d}'.format(len(unfinished_tasks)))
    #        if duration:
    #            _display('Total time: {:.1f} seconds'.format(duration))

    def _worker(self, task, task_queue):
        """ Handles processing of the tasks.
        """
        with _thread_lock:
            task.process_number = self.counter
            self.counter += 1
        if task.output:
            if not task.has_error and task.processtime > 0:
                if not os.path.isfile(task.logfile):
                    task.logfile = ""
                task_queue.put(task)
                return
        try:
            if not os.path.exists(task.folder):
                task.add_error('Could not find folder: {}'.format(task.folder))
                task.logfile = ""
            else:
                tmp_kwargs = dict(mode='a+', prefix=self.logfile_prefix,
                                  suffix='.log', dir=task.folder, delete=False)
                with NamedTemporaryFile(**tmp_kwargs) as logfile:
                    logfile.write('########### MACRO #############\n')
                    logfile.write("\n".join(task.macro))
                    logfile.write('\n\n######### OUTPUT LOG ##########')
                    logfile.flush()
                    task.logfile = logfile.name
                    starttime = time.clock()
                    exe_args = dict(macro=task.macro,
                                    logfile=logfile,
                                    anybodycon_path=self.anybodycon_path,
                                    timeout=self.timeout,
                                    keep_macrofile=self.keep_logfiles,
                                    env=self.env)
                    retcode = _execute_anybodycon(**exe_args)
                    endtime = time.clock()
                    logfile.seek(0)
                    if retcode == _KILLED_BY_ANYPYTOOLS:
                        task.processtime = 0
                        return
                    task.processtime = endtime - starttime
                    task.output = parse_anybodycon_output(
                        logfile.read(),
                        self.ignore_errors,
                        self.warnings_to_include)
        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
            task.add_error(str(exc_type) + '\n' + str(fname) + '\n' + str(exc_tb.tb_lineno))
            logger.debug(str(e))
        finally:
            if not self.keep_logfiles and not task.has_error:
                try:
                    silentremove(logfile.name)
                    task.logfile = ""
                except OSError as e:
                    pass  # Ignore if AnyBody has not released the log file.
            task_queue.put(task)

    def _schedule_processes(self, tasklist, _worker):
        # Reset the global flag that allows
        global _stop_all_processes
        _subprocess_container.stop_all = False
        # Make a shallow copy of the task list,
        # so we don't mess with the callers list.
        tasklist = copy.copy(tasklist)
        number_tasks = len(tasklist)
        if number_tasks == 0:
            totaltime = 0
            return totaltime
        use_threading = (number_tasks > 1 and self.num_processes > 1)
        starttime = time.clock()
        task_queue = Queue()
        pbar = _ProgressBar(number_tasks, self.silent)
        pbar.animate(0)
        processed_tasks = []
        n_errors = 0
        threads = []
        try:
            # run while there is still threads, tasks or stuff in the queue
            # to process
            while threads or tasklist or task_queue.qsize():
                # if we aren't using all the processors AND there is still
                # data left to compute, then spawn another thread
                if (len(threads) < self.num_processes) and tasklist:
                    if use_threading:
                        t = Thread(target=_worker,
                                   args=tuple([tasklist.pop(0), task_queue]))
                        t.daemon = True
                        t.start()
                        threads.append(t)
                    else:
                        _worker(tasklist.pop(0), task_queue)
                else:
                    # In the case that we have the maximum number
                    # of running threads or we run out tasks.
                    # Check if any of them are done
                    for thread in threads:
                        if not thread.isAlive():
                            threads.remove(thread)
                while task_queue.qsize():
                    task = task_queue.get()
                    if task.has_error:
                        n_errors += 1
                    self.summery.task_summery(task)
                    processed_tasks.append(task)
                    pbar.animate(len(processed_tasks), n_errors)

                time.sleep(0.01)
        except KeyboardInterrupt:
            _display('Processing interrupted')
            _subprocess_container.stop_all = True
            # Add a small delay here. It allows the user to press ctrl-c twice
            # to escape this try-catch. This is usefull when if the code is
            # run in an outer loop which we want to excape as well.
            time.sleep(1)
        totaltime = time.clock() - starttime
        return totaltime

    def cleanup_logfiles(self, tasklist):
        for task in tasklist:
            try:
                if not self.keep_logfiles and not task.has_error:
                    if task.logfile:
                        silentremove(task.logfile)
                        task.logfile = ""
                if task.processtime <= 0 and task.logfile:
                    silentremove(task.logfile)
                    task.logfile = ""
            except OSError as e:
                logger.debug('Could not remove '
                             '{} {}'.format(task.logfile, str(e)))
            if not self.keep_logfiles:
                try:
                    macrofile = task.logfile.replace('.log', '.anymcr')
                    silentremove(macrofile)
                except OSError as e:
                    logger.debug('Could not removing '
                                 '{} {}'.format(macrofile, str(e)))


class _ProgressBar:
    def __init__(self, iterations, silent=False):
        self.silent = silent
        self.iterations = iterations
        self.prog_bar = '[]'
        self.fill_char = '*'
        self.width = 40
        if run_from_ipython() and not self.silent:
            self.bar_widget = ipywidgets.IntProgress(
                min=0, max=iterations, value=0, bar_style='')
            self.bar_description = ipywidgets.Label('')
            box = ipywidgets.HBox([self.bar_description, self.bar_widget])
            box.layout.align_items = 'center'
            display(box)

    def animate(self, val, failed=0):
        if self.silent:
            return
        if run_from_ipython():
            self._widget_animate(val, failed)
        else:
            self._ascii_animate(val, failed)

    def _widget_animate(self, val, failed):
        self.bar_widget.value = val
        self.bar_description.value = '%d of %s' % (val, self.iterations)
        if failed > 0:
            self.bar_widget.bar_style = 'danger'
        elif val == self.iterations:
            self.bar_widget.bar_style = 'success'
            
            
    def _ascii_animate(self, val, failed):
        self.__update_amount((val / float(self.iterations)) * 100.0)
        self.prog_bar += '  %d of %s complete' % (val,self.iterations)
        if failed == 1:
            self.prog_bar += ' ({0} Error)'.format(failed)
        elif failed > 1:
            self.prog_bar += ' ({0} Errors)'.format(failed)
        print('\r', end="")
        print(self.prog_bar, end="")
        sys.stdout.flush()

    def __update_amount(self, new_amount):
        percent_done = int(round((new_amount / 100.0) * 100.0))
        all_full = self.width - 2
        num_hashes = int(round((percent_done / 100.0) * all_full))
        self.prog_bar = ('[' +
                         self.fill_char * num_hashes +
                         ' ' * (all_full - num_hashes) +
                         ']')
        pct_place = int(len(self.prog_bar) / 2) - len(str(percent_done))
        pct_string = '%d%%' % percent_done
        self.prog_bar = self.prog_bar[0:pct_place] + \
                        (pct_string + self.prog_bar[pct_place + len(pct_string):])


if __name__ == '__main__':
    pass



# -*- coding: utf-8 -*-
"""
Created on Mon Mar 23 21:14:59 2015

@author: Morten
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *
from past.builtins import basestring as string_types


import sys
import logging
from copy import deepcopy
from collections import MutableSequence

import numpy as np


from scipy.stats import distributions

def is_python35():
    return (sys.version_info[0]==3 and sys.version_info[1]>=5)


if sys.platform.startswith("win") and not is_python35():
    # This is a horrible hack to work around a bug in
    # scipy http://stackoverflow.com/questions/15457786/ctrl-c-crashes-python-after-importing-scipy-stats
    try:
        import thread  #, imp, ctypes, os
    except ImportError:
        import _thread as thread
    import win32api

    def handler(sig, hook=thread.interrupt_main):
        hook()
        return 1

    win32api.SetConsoleCtrlHandler(handler, 1)

logger = logging.getLogger('abt.anypytools')

#pprint is used in the doc tests
from .tools import define2str, path2str, array2anyscript, pprint
from .tools import Py3kPrettyPrinter

__all__ = ['MacroCommand', 'Load', 'SetValue', 'SetValue_random', 'Dump',
           'SaveDesign', 'LoadDesign', 'SaveValues', 'LoadValues', 'UpdateValues',
           'OperationRun']


def _isgenerator(x):
    return isinstance(x, types.GeneratorType)

def _batch(iterable, n = 1):
   l = len(iterable)
   for ndx in range(0, l, n):
       yield iterable[ndx:min(ndx+n, l)]


class MacroCommand(object):
    """ Class for custom macro commands

    This class also serves as base class for other macro commands.

    Examples
    --------
    >>> mc = MacroCommand('operation "Main.MyStudy.InverseDynamics"') )
    >>> print(mc)
    operation "Main.MyStudy.InverseDynamics"

    """
    def __init__(self, command):
        if not isinstance(command, list):
            self.cmd = [command]
        else:
            self.cmd = command

    def __repr__(self):
        return self.get_macro(0)

    def get_macro(self, index, **kwarg):
        return '\n'.join(self.cmd)


class Load(MacroCommand):
    """ Create a load macro command

    Parameters
    ----------
    filename : str
        Path of the file to load
    defs : dict
        Dictionary of defines statements to set during load
    paths : dict
        Dictionary of path staements to set during load

    Examples
    --------
    >>> Load('model.main.any')
    load "model.main.any"

    >>> paths = {'DATA':'c:/MyModel/Data'}
    >>> defines = {'EXCLUDE_ARMS':None, 'N_STEP':20}
    >>> Load('c:/MyModel/model.main.any', defines, paths)
    load "c:/MyModel/model.main.any" -def EXCLUDE_ARMS="" -def N_STEP="20" -p DATA=---"c:/MyModel/Data"

    >>> mcr = AnyMacro( Load('model_{id}.main.any'), counter_token = '{id}')
    >>> mcr.create_macros(3)
    [[u'load "model_0.main.any"'],
     [u'load "model_1.main.any"'],
     [u'load "model_2.main.any"']]

    """
    def __init__(self, filename, defs={}, paths={}):
        self.filename = filename
        self.defs = defs.copy()
        self.paths = paths.copy()

    def get_macro(self, index, **kwarg):
        cmd = ['load "{}"'.format(self.filename)]

        for key in sorted(self.defs):
            value = self.defs[key]
            cmd.append(define2str(key, value))
        for key in sorted(self.paths):
            value = self.paths[key]
            cmd.append(path2str(key, value))

        return ' '.join(cmd)


class SetValue(MacroCommand):
    """ Create 'Set Value' classoperation macro command.

    Parameters
    ----------
    var : string
        An AnyScript variable.
    value : number or list of number
        A value or list of values to assign to the AnyScript variable.

    Examples
    ---------
    Set a single values:

    >>> SetValue('Main.Study.myvar1', 23.1)
    classoperation Main.Study.myvar1 "Set Value" --value="23.1"

    >>> SetValue('Main.Study.myvar2', np.array([2,3,4]))
    classoperation Main.Study.myvar2 "Set Value" --value="{2,3,4}"

    Set variable across different macros

    >>> SetValue('Main.Study.myvar1',[1,2,3])
    classoperation Main.Study.myvar1 "Set Value" --value="1"
    classoperation Main.Study.myvar1 "Set Value" --value="2"
    classoperation Main.Study.myvar1 "Set Value" --value="3"

    Here is a elaborate example:

    >>> mg = AnyMacro( Load('MyModel.main.any'),
                     SetValue('Main.Study.myvar1',[1,2]),
                     OperationRun('Main.Study.InverseDynamics') )
    >>> mg.number_of_macros = 2
    [['load "MyModel.main.any"',
      'classoperation Main.Study.myvar1 "Set Value" --value="1"',
      'operation Main.Study.InverseDynamics',
      'run'],
     ['load "MyModel.main.any"',
      'classoperation Main.Study.myvar1 "Set Value" --value="2"',
      'operation Main.Study.InverseDynamics',
      'run']]

    If we generate more macros that there are values then the values are just
    repeated

    >>> mg.number_of_macros = 4
    >>> mg
    [['load "MyModel.main.any"',
      'classoperation Main.Study.myvar1 "Set Value" --value="1"',
      'operation Main.Study.InverseDynamics',
      'run'],
     ['load "MyModel.main.any"',
      'classoperation Main.Study.myvar1 "Set Value" --value="2"',
      'operation Main.Study.InverseDynamics',
      'run'],
     ['load "MyModel.main.any"',
      'classoperation Main.Study.myvar1 "Set Value" --value="1"',
      'operation Main.Study.InverseDynamics',
      'run'],
     ['load "MyModel.main.any"',
      'classoperation Main.Study.myvar1 "Set Value" --value="2"',
      'operation Main.Study.InverseDynamics',
      'run']]

    """

    def __init__(self, var, value):
        self.var = var
        self.value = value

    def __repr__(self):
        if isinstance(self.value, list):
            n_elem = len(self.value)
        else:
            n_elem = 1

        return '\n'.join([self.get_macro(i) for i in range(n_elem)])


    def get_macro(self, index, **kwarg):
        if isinstance(self.value, list):
            val = self.value[index % len(self.value)]
        else:
            val = self.value
        return self.format_macro(val)

    def format_macro(self, val):
        if isinstance(val, np.ndarray):
            val = array2anyscript(val)
        elif isinstance(val, float):
            val = '{:.12g}'.format(val)
        elif isinstance(val, int):
            val = '{:d}'.format(val)
        return 'classoperation {0} "Set Value" --value="{1}"'.format(self.var,
                                                                     val)


class SetValue_random(SetValue):
    """ Create a 'Set Value' macro command where the value is chosen from a
    distibution in scipy.stats.distributions.

    Parameters
    ----------
    var : str
        An AnyScript variable.
    frozen_distribution : <scipy.stats.distributions.rv_frozen>
        A frozen distribution from scipy.stats.distributions
    default_lower_tail_probability : float
        The lower tail probability of the default value. Defaults to 0.5 which
        is the mean value.


    Examples
    --------
    Creating normal macros will use the default values. Usually the 50 percentile
    (mean values).

    >>> np.random.seed(1)
    >>> from scipy.stats.distributions import logistic, norm
    >>> log_dist = logistic( loc= [1,3,4],scale = [0.1,0.5,1] )
    >>> norm_dist = norm( loc= [0,0,0],scale = [0.1,0.5,1] )
    >>> cmd = [SetValue_random('Main.MyVar1', log_dist), SetValue_random('Main.MyVar2', norm_dist) ]
    >>> mg = AnyMacro(cmd, number_of_macros = 3)
    >>> mg
    [['classoperation Main.MyVar1 "Set Value" --value="{0,0,0}"',
      'classoperation Main.MyVar2 "Set Value" --value="{1,3,4}"'],
     ['classoperation Main.MyVar1 "Set Value" --value="{0,0,0}"',
      'classoperation Main.MyVar2 "Set Value" --value="{1,3,4}"'],
     ['classoperation Main.MyVar1 "Set Value" --value="{0,0,0}"',
      'classoperation Main.MyVar2 "Set Value" --value="{1,3,4}"']]

    Values can be sampled randomly in a 'Monte Carlo' fashion. Note that the
    first value is still the defaut mean values

    >>> mg.create_macros_MonteCarlo()
    [['classoperation Main.MyVar1 "Set Value" --value="{0,0,0}"',
      'classoperation Main.MyVar2 "Set Value" --value="{1,3,4}"'],
     ['classoperation Main.MyVar1 "Set Value" --value="{-0.0209517840916,0.291902872134,-3.68494766954}"',
      'classoperation Main.MyVar2 "Set Value" --value="{0.916378512121,2.11986245798,1.71459079162}"'],
     ['classoperation Main.MyVar1 "Set Value" --value="{-0.0891762169545,-0.198666812922,-0.261723075945}"',
      'classoperation Main.MyVar2 "Set Value" --value="{1.01555799978,2.83695956934,4.7778636562}"']]

    Values can also be sampled with a Latin Hyper Cube sampler.

    >>> mg.create_macros_LHS()
    [['classoperation Main.MyVar1 "Set Value" --value="{0.0928967116493,-0.591418725401,-0.484696993931}"',
      'classoperation Main.MyVar2 "Set Value" --value="{1.01049414425,2.46211329129,5.73806916203}"'],
     ['classoperation Main.MyVar1 "Set Value" --value="{-0.0166741228961,0.707722119582,-0.294180629253}"',
      'classoperation Main.MyVar2 "Set Value" --value="{1.11326829265,2.66016732923,4.28054911097}"'],
     ['classoperation Main.MyVar1 "Set Value" --value="{-0.20265197275,0.114947152258,0.924796936287}"',
      'classoperation Main.MyVar2 "Set Value" --value="{0.806864877696,4.4114188826,2.93941843565}"']]

   """

    def __init__(self, var, frozen_distribution, default_lower_tail_probability = 0.5):
        self.var = var
        if not isinstance(frozen_distribution, distributions.rv_frozen):
            raise TypeError(
                'frozen_distribution must be frozen distribtuion from \
                            scipy.stats.distributions')

        self.rv = frozen_distribution

        self.meanvalue = frozen_distribution.mean()
        if isinstance(self.meanvalue, np.ndarray):
            self.shape = self.meanvalue.shape
            self.n_factors = np.prod(self.shape)
            self.default_lower_tail_probability = default_lower_tail_probability * np.ones(self.shape)
        else:
            self.shape = None
            self.n_factors = 1
            self.default_lower_tail_probability = default_lower_tail_probability



    def __repr__(self):
        return self.get_macro(0)


    def get_macro(self, index, lower_tail_probability=None, **kwarg):
        if lower_tail_probability is None:
            lower_tail_probability = self.default_lower_tail_probability

        if self.shape is None and isinstance(lower_tail_probability, np.ndarray):
                lower_tail_probability = lower_tail_probability[0]

        val = self.rv.ppf(lower_tail_probability).reshape(self.shape)
        # Replace any nan from ppf function with actual sampled values.
        if np.any(np.isnan(val)):
            val[np.isnan(val)] = self.rv.rvs()[np.isnan(val)]

        return self.format_macro(val)


class Dump(MacroCommand):
    """ Create a Dump classoperation macro command.

    Parameters
    ----------
    var : str or list of str
        The anyscript values to create a 'Dump' macro command for
    include_in_macro: int or list of int
        Specifices in which macros [0,1,2....NumberOfMacros] to include the
        dump command.
        If None, the command is included in all macros.

    Examples
    ---------
    >>> Dump('Main.Study.myvar1')
    classoperation Main.Study.myvar1 "Dump"

    Only include the dump command in the two first macro

    >>> mg = AnyMacro(number_of_macros = 5)
    >>> mg.append(Load('mymodel.main.any'))
    >>> mg.append(Dump('Main.Study.myvar1', include_in_macro = [0,1]))
    >>> mg
    [['load "mymodel.main.any"', 'classoperation Main.Study.myvar1 "Dump"'],
     ['load "mymodel.main.any"', 'classoperation Main.Study.myvar1 "Dump"'],
     ['load "mymodel.main.any"'],
     ['load "mymodel.main.any"'],
     ['load "mymodel.main.any"']]
    """

    def __init__(self, var, include_in_macro = None):
        if not isinstance(var, list):
            self.var_list = [var]
        else:
            self.var_list = var
        self._include_in_macro = include_in_macro

    def get_macro(self, index, **kwarg):
        cmd = []
        if self._include_in_macro is None or index in self._include_in_macro:
            for var in self.var_list:
                cmd.append('classoperation {0} "Dump"'.format(var))
        return '\n'.join(cmd)


class SaveDesign(MacroCommand):
    """ Create a Save Design classoperation macro command.

    Parameters
    ----------
    operation : str
        The AnyScript operation
    filename : str
        The filename in which to save the design

    Examples
    --------
    >>> SaveDesign('Main.MyStudy.Kinematics', 'c:/design.txt')
    classoperation Main.MyStudy.Kinematics "Save design" --file="c:/design.txt"
    """

    def __init__(self, operation, filename):
        self.filename = filename
        self.operation = operation

    def get_macro(self, index, **kwarg):
        return 'classoperation {} "Save design" --file="{}"'.format(
            self.operation, self.filename)


class LoadDesign(MacroCommand):
    """ Create a Load Design classoperation macro command.

    Parameters
    ----------
    operation : str
        The AnyScript operation
    filename : str
        The file in which to load the design from

    Examples
    --------
    >>> LoadDesign('Main.MyStudy.Kinematics', 'c:/design.txt')
    classoperation Main.MyStudy.Kinematics "Load design" --file="c:/design.txt"
    """
    def __init__(self, operation, filename):
        self.filename = filename
        self.operation = operation

    def get_macro(self, index, **kwarg):
        return 'classoperation {} "Load design" --file="{}"'.format(
            self.operation, self.filename)


class SaveValues(MacroCommand):
    """ Create a Save Values classoperation macro command.

    Parameters
    ----------
    filename : str
        The anyset file to save the values to

    Examples
    --------
    >>> SaveValues('c:/values.anyset')
    classoperation Main "Save Values" --file="c:/values.anyset"
    """
    def __init__(self, filename):
        self.filename = filename

    def get_macro(self, index, **kwarg):
        return 'classoperation Main "Save Values" --file="{}"'.format(
            self.filename)


class LoadValues(MacroCommand):
    """ Create a Load Values classoperation macro command.

    Parameters
    ----------
    filename : str
        The anyset file to load the values from.

    Examples
    --------
    >>> LoadValues('c:/values.anyset')
    classoperation Main "Load Values" --file="c:/values.anyset"
    """
    def __init__(self, filename):
        self.filename = filename

    def get_macro(self, index, **kwarg):
        return 'classoperation Main "Load Values" --file="{}"'.format(
            self.filename)


class UpdateValues(MacroCommand):
    """ Create an 'Update Values' classoperation macro command.

    Examples
    --------
    >>> UpdateValues()
    classoperation Main "Update Values"
    """
    def __init__(self):
        pass

    def get_macro(self, index, **kwarg):
        return 'classoperation Main "Update Values"'


class OperationRun(MacroCommand):
    """ Create a macro command to select and run an operation

    Examples
    --------
    >>> OperationRun('Main.MyStudy.Kinematics')
    >>> pprint( mg.generate_macros())
    operation Main.MyStudy.Kinematics
    run

    >>> mg = AnyMacro(Load('my_model.main.any'),
                   OperationRun('Main.MyStudy.Kinematics'))
    >>> mg.create_macros()
    [[u'load "my_model.main.any"', u'operation Main.MyStudy.Kinematics', u'run']]
    """
    def __init__(self, operation):
        self.operation = operation

    def get_macro(self, index, **kwarg):
        return 'operation {}'.format(self.operation) + '\n' + 'run'


class AnyMacro(MutableSequence):
    """
    Class to generate multiple macros.

    Use the class to contruct multiple anyscript macros from a single macro.

    Parameters
    ----------
    maco_commands : list or positional arguments
        A list of macro command classes. These may also be given a arbitrary number
        of positional arguments.
    counter_token : str
        A string in the macro commands which will be replace with a counter
        for every macro generated
    number_of_macros : int
        The default number of macros which are created. This can be overruled when
        calling the AnyMacro.create_macros() function. Defults to 1.
    seed : int or array_like
        A constant seed value used which will be used every time macros are created.
        Must be convertable to 32 bit unsigned integers.

    Returns
    -------
    An AnyMacro instance
        A Macro object for constructing the macro.


    Examples
    --------
    >>> mg = AnyMacro(number_of_macros = 22)
    >>> mg.append( Load('c:/MyModel/model.main.any', defs = {}, paths = {} ) )
    >>> mg.append( SetValue('Main.myvar', 12.3)  )
    >>> mg.extend( [RunOperation('Main.Study.Kinematics'),
                     Dump('Main.Study.Output.MyVar') ] )
    >>> mg
    [['load "c:/MyModel/model.main.any"',
      'classoperation Main.myvar "Set Value" --value="12.3"',
      'operation Main.Study.Kinematics',
      'run',
      'classoperation Main.Study.Output.MyVar "Dump"'],
     ['load "c:/MyModel/model.main.any"',
      'classoperation Main.myvar "Set Value" --value="12.3"',
      'operation Main.Study.Kinematics',
      'run',
      'classoperation Main.Study.Output.MyVar "Dump"']]


    >>> mg = AnyMacro( number_of_macros = 3, counter_token='{COUNTER}' )
    >>> mg.append(Load('c:/MyModel/model_{COUNTER}_.main.any') )
    >>> pprint(mg.create_macros())
    [['load "c:/MyModel/model_0_.main.any"'],
     ['load "c:/MyModel/model_1_.main.any"'],
     ['load "c:/MyModel/model_2_.main.any"']]



    """

    def __init__(self, *args,
                 **kwargs):  # number_of_macros=1, counter_token = None ):
        super(MutableSequence, self).__init__()
        if len(args) == 0:
            self._list = list()
        else:
            if len(args) == 1 and isinstance(args[0], list):
                args = args[0]
            for arg in args:
                if not isinstance(arg, MacroCommand):
                    raise ValueError(
                        'Argument must be valid macro command classes')
            self._list = list(args)

        self.seed = kwargs.pop("seed", None)

        self.number_of_macros = kwargs.pop("number_of_macros", 1)
        self.counter_token = kwargs.pop("counter_token", None)

    def __mul__(self, other):
        if not isinstance(other, (int, float)):
            raise NotImplementedError('operator must be integer')
        mg = deepcopy(self)
        mg.number_of_macros = mg.number_of_macros*round(other)
        return mg

    __rmul__ = __mul__

    def __len__(self):
        return len(self._list)

    def __getitem__(self, ii):
        return self._list[ii]

    def __delitem__(self, ii):
        del self._list[ii]

    def __setitem__(self, ii, val):
        self._list[ii] = val
        return self._list[ii]

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        printer = Py3kPrettyPrinter(width=80)
        return printer.pformat(self.create_macros())

    def insert(self, ii, val):
        self._list.insert(ii, val)

    def append(self, val):
        list_idx = len(self._list)
        self.insert(list_idx, val)




    def create_macros(self, number_of_macros=None, batch_size = None):
        if batch_size is not None:
            return _batch(self.create_macros(number_of_macros), n = batch_size)

        if number_of_macros is None:
            number_of_macros = self.number_of_macros
        if self.seed is not None:
            np.random.seed(self.seed)
        macro_list = []
        for macro_idx in range(number_of_macros):
            macro = []
            for elem in self:
                mcr = elem.get_macro(macro_idx)
                if self.counter_token:
                    mcr = mcr.replace(self.counter_token, str(macro_idx))
                if mcr is not '':
                    macro.extend(mcr.split('\n'))
            macro_list.append(macro)
        return macro_list


    def create_macros_MonteCarlo(self, number_of_macros=None, batch_size = None):

        if batch_size is not None:
            return _batch(self.create_macros_MonteCarlo(number_of_macros), batch_size)

        if number_of_macros is None:
            number_of_macros = self.number_of_macros

        if self.seed is not None:
            np.random.seed(self.seed)


        macro_list = []
        for macro_idx in range(number_of_macros):
            macro = []
            lhs_idx = 0
            for elem in self:
                if isinstance(elem, SetValue_random):
                    if macro_idx == 0:
                        lower_tail_probability = None # First macro get the default values
                    else:
                        lower_tail_probability = np.random.random(elem.n_factors)
                    mcr = elem.get_macro(macro_idx, lower_tail_probability)
                    lhs_idx += elem.n_factors
                else:
                    mcr = elem.get_macro(macro_idx)

                if self.counter_token:
                    mcr = mcr.replace(self.counter_token, str(macro_idx))
                if mcr is not '':
                    macro.extend(mcr.split('\n'))
            macro_list.append(macro)
        return macro_list




    def create_macros_LHS(self,
                          number_of_macros=None,
                          criterion=None,
                          iterations=None,
                          batch_size = None,
                          append_default=False):
        try:
            import pyDOE
        except ImportError:
            raise ImportError(
                'The pyDOE package must be install to use this class')

        if batch_size is not None:
            return _batch(self.create_macros_LHS(number_of_macros,
                                                 criterion,
                                                 iterations), n = batch_size)

        if number_of_macros is None:
            number_of_macros = self.number_of_macros

        if self.seed is not None:
            np.random.seed(self.seed)

        factors = sum([e.n_factors for e in self
                       if isinstance(e, SetValue_random)])
        lhs_matrix = pyDOE.lhs(factors, number_of_macros,
                               criterion=criterion,
                               iterations=iterations)

        macro_list = []

        for macro_idx in range(number_of_macros):
            macro = []
            lhs_idx = 0
            for elem in self:
                if isinstance(elem, SetValue_random):
                    lhs_val = lhs_matrix[macro_idx, lhs_idx:lhs_idx +
                                         elem.n_factors]
                    mcr = elem.get_macro(macro_idx,
                                         lower_tail_probability=lhs_val)
                    lhs_idx += elem.n_factors
                else:
                    mcr = elem.get_macro(macro_idx)

                if self.counter_token:
                    mcr = mcr.replace(self.counter_token, str(macro_idx))
                if len(mcr) > 0:
                    macro.extend(mcr.split('\n'))
            macro_list.append(macro)
        if append_default is not None:
            macro = []
            for elem in self:
                mcr = elem.get_macro(number_of_macros)
                if self.counter_token:
                   mcr = mcr.replace(self.counter_token, str(number_of_macros))
                if len(mcr) > 0:
                    macro.extend(mcr.split('\n'))
            macro_list.append(macro)
        return macro_list


if __name__ == '__main__':

    #    mg = PertubationMacroGenerator()
    #    mg.add_set_value_designvar('test',1)
    #    mg.add_set_value_designvar('test',2)
    #
    #    pprint(mg.generate_macros())
    #
    import pytest
    pytest.main(str('generate_macros.py --doctest-modules'))



# -*- coding: utf-8 -*-
"""
Created on Wed Apr 29 08:21:46 2015

@author: mel
"""

from .macroutils import ( MacroCommand, Load, SetValue, SetValue_random,
                          Dump, SaveDesign, LoadDesign, SaveValues,
                          LoadValues, UpdateValues, OperationRun )

__all__ = ['MacroCommand', 'Load', 'SetValue', 'SetValue_random',
           'Dump', 'SaveDesign', 'LoadDesign', 'SaveValues',
           'LoadValues', 'UpdateValues', 'OperationRun']




# -*- coding: utf-8 -*-
"""
Created on Mon Jan 16 11:40:42 2012

@author: mel
"""
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import *

import logging

import h5py
logger = logging.getLogger('abt.anypytools')

def _follow_reftarget(elem):
    completename = elem.attrs['CompleteName']
    completename = completename.replace(b'.',b'/')
    reftarget = elem.attrs['RefTarget'].replace(b'.',b'/')
    prefix = completename[:-len(elem.name)]
    h5target = reftarget[len(prefix):]
    elem = elem.file[h5target]
    return elem


def _check_input_path(path):
    if not "/" in path:
        # path does not have traditional h5 format.
        if path.startswith('Main.') and 'Output' in path:
            path = '/Output' + path.split('Output')[-1]
        path = path.replace('.', '/')
    return path


class File(h5py.File):
    def __init__(self, *args, **kwargs):
         super(File, self).__init__(*args, **kwargs)
         self.wrapped = True

    def __getitem__(self,path):
        path = _check_input_path(path)
        try:
            elem = super(File, self).file[path]
            if isinstance(elem, h5py.Group) and not len(elem.keys()):
                if 'RefTarget' in elem.attrs:
                    elem = _follow_reftarget(elem)
        except KeyError:
            elem = super(type(self),self)
            levels = path.strip('/').split('/')
            for level in levels:
                if elem.__contains__(level):
                    elem = elem.__getitem__(level)
                else:
                    try:
                        elem = _follow_reftarget(elem)
                        elem = elem.__getitem__(level)
                    except:
                        raise KeyError('Entry not found: '+path    )
        if isinstance(elem, h5py.Group):
            return Group(elem.id)
        elif isinstance(elem, h5py.Dataset):
            return Dataset(elem.id)
        elif isinstance(elem,h5py.File):
            return File(elem.id)

    @property
    def file(self):
        id = super(File,self).file.id
        return File(id)

    @property
    def parent(self):
        id = super(File,self).parent.id
        return Group(id)


    def __contains__(self, name):
        """ Test if a member name exists """
        if super(File, self).__contains__(name):
            return True
        else:
            try:
                self.__getitem__(name)
                return True
            except KeyError:
                pass
        return False
        
File.__doc__ = h5py.File.__doc__

class Group(h5py.Group):
    def __init__(self,arg):
        super(Group, self).__init__(arg)
        self.wrapped = True

    def __getitem__(self,path):
        path = _check_input_path(path)
        try:
            elem = super(Group, self).__getitem__(path)
            if isinstance(elem, h5py.Group) and not len(elem.keys()):
                if 'RefTarget' in elem.attrs:
                    elem = _follow_reftarget(elem)
        except KeyError:
            elem = super(type(self),self)
            levels = path.strip('/').split('/')
            for level in levels:
                if elem.__contains__(level):
                    elem = elem.__getitem__(level)
                else:
                    try:
                        elem = _follow_reftarget(elem)
                        elem = elem.__getitem__(level)
                    except:
                        raise KeyError('Entry not found: '+path    )
        if isinstance(elem, h5py.Group):
            return Group(elem.id)
        elif isinstance(elem, h5py.Dataset):
            return Dataset(elem.id)
        elif isinstance(elem,h5py.File):
            return File(elem.id)

    @property
    def file(self):
        id = super(Group,self).file.id
        return File(id)

    @property
    def parent(self):
        id = super(Group,self).parent.id
        return Group(id)

    def __contains__(self, name):
        """ Test if a member name exists """
        if super(Group, self).__contains__(name):
            return True
        else:
            try:
                self.__getitem__(name)
                return True
            except KeyError:
                pass
        return False

Group.__doc__ = h5py.Group.__doc__
        
        
class Dataset(h5py.Dataset):
    def __init__(self,arg):
        super(Dataset, self).__init__(arg)
        self.wrapped = True

    @property
    def file(self):
        id = super(Dataset,self).file.id
        return File(id)

    @property
    def parent(self):
        id = super(Dataset,self).parent.id
        return Group(id)

Dataset.__doc__ = h5py.Dataset.__doc__
        
        
if __name__ == '__main__':
#    for data,header,filename in csv_trial_data('C:\Users\mel\SMIModelOutput', DEBUG= True):
#        print header

    outvars = ['/Output/Validation/EMG']

    for data, filename in h5_trial_data(outvars):
        print( data.keys() )


# -*- coding: utf-8 -*-
"""
Created on Mon Sep  1 12:44:36 2014

@author: Morten
"""
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import *

import os
import ast
import shutil
import textwrap
import itertools
import contextlib
import subprocess 
from traceback import format_list, extract_tb

import pytest

from anypytools import AnyPyProcess, macro_commands
from anypytools.tools import get_anybodycon_path, replace_bm_constants
from anypytools.generate_macros import MacroGenerator


def _limited_traceback(excinfo):
    """ Return a formatted traceback with all the stack
        from this frame (i.e __file__) up removed
    """
    tb = extract_tb(excinfo.tb)
    try:
        idx = [__file__ in e for e in tb].index(True)
        return format_list(tb[idx+1:])
    except ValueError:
        return format_list(tb)

        
def _read_header(path):
    code = ''
    with open(path) as f:
        for line in f.readlines():
            if line.startswith('//'):
                line = line.strip('//')
                code += line
            else:
                break
    return code
            
def _exec_header(header):
    ns = {}
    try:
        exec(header, globals(), ns)
    except SyntaxError:
        pass
    if len(ns) == 0:
        try:
            ns['define'] = ast.literal_eval(header)
        except SyntaxError:
            pass
    return ns


@contextlib.contextmanager        
def change_dir(path):
    prev_cwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(prev_cwd)    
     

def pytest_collect_file(parent, path):
    if path.ext.lower() == ".any" and path.basename.lower().startswith("test_"):
            return AnyFile(path, parent)
    elif parent.config.getoption("--collect-main-files"):
        if path.basename.lower().endswith('main.any'):
            return AnyFile(path, parent)
      
def _format_switches(defs):
    if isinstance(defs, dict):
        defs = [defs]
    elif isinstance(defs, tuple):
        combinations = list(itertools.product(*defs))
        defs = []
        for elem in combinations:
            defs.append({k:v for d in elem for k,v in d.items()})
    elif isinstance(defs, list):
        pass
    else:
        defs = [{}]
    if len(defs) == 0:
        defs = [{}]
    return defs
            
def _as_absolute_paths(d, start): 
    return {k: os.path.abspath(os.path.relpath(v,start)) for k,v in d.items()}

    
class AnyFile(pytest.File):
    def collect(self):
        # Collect define statements from the header
        #import pdb;pdb.set_trace()
        header = _read_header(self.fspath.strpath)
        ns = _exec_header(header)
        def_list = _format_switches(ns.get('define', {}))
        def_list = [replace_bm_constants(d) for d in def_list]
        path_list = _format_switches(ns.get('path', {}))
        combinations = itertools.product(def_list, path_list)
        ignore_errors = ns.get('ignore_errors',[]) 
        if not isinstance(ignore_errors,list):
            raise TypeError('ignore_errors must be a list')  
        expect_errors = ns.get('expect_errors',[]) 
        if not isinstance(expect_errors,list):
            raise TypeError('expect_errors must be a list')   
        # Run though the defines an create a test case for each
        for i, (defs, paths) in enumerate(combinations):
            name = '{}_{}'.format(self.fspath.basename,i) 
            if isinstance(defs, dict) and isinstance(paths, dict):
                yield AnyItem(
                    name = name, 
                    parent = self, 
                    defs=defs,
                    paths=paths,
                    ignore_errors = ignore_errors,
                    expect_errors = expect_errors
                )
            else:
                raise ValueError('Malformed input: ', header)

                
class AnyItem(pytest.Item):
    def __init__(self, name, parent, defs, paths, ignore_errors=None, expect_errors=None):
        super().__init__(name, parent)
        self.defs = defs
        self.defs['TEST_NAME'] = '"{}"'.format(name)
        self.paths = _as_absolute_paths(paths, self.fspath.dirname)
        self.name = name
        self.expect_errors = expect_errors
        self.errors = []
        self.macro = [macro_commands.Load(self.fspath.strpath,
                                          self.defs, self.paths)]
        self.macro_file = None     
        if not self.config.getoption("--only-load"):
            self.macro.append(macro_commands.OperationRun('Main.RunTest') )
            user_macro =  self.config.getoption("--run-macro")
            if user_macro is not None:
                self.macro.append(macro_commands.OperationRun(user_macro))
        self.apt_opts = {
            'return_task_info': True,
            'silent': True,
            'anybodycon_path': self.config.getoption("--anybodycon"),
            'ignore_errors': ignore_errors,
            }
            

    def runtest(self):
        tmpdir = self.config._tmpdirhandler.mktemp(self.name)
        with change_dir(tmpdir.strpath):
            app = AnyPyProcess(**self.apt_opts)
            result = app.start_macro(self.macro)[0]
            self.app = app
        # Ignore error due to missing Main.RunTest
        if 'ERROR' in result:
            for i, e in enumerate(result['ERROR']):
                if 'Error : Main.RunTest : Unresolved object' in e:
                    del result['ERROR'][i]
                    break
        # Check that the expected errors are present
        if self.expect_errors is not None:
            if 'ERROR' in result:
                # Remove any failures due to RunTest not working on failed models
                for i, e in enumerate(result['ERROR']):
                    if 'Main.RunTest : Select Operation is not expected' in e:
                        del result['ERROR'][i]
                        break
                error_list = result['ERROR']
            else:
                error_list = []
            for xerr in self.expect_errors:
                xerr_found = False
                for i, error in enumerate(error_list):
                    if xerr in error:
                        xerr_found = True
                        del error_list[i]
                if not xerr_found:
                    self.errors.append('TEST ERROR: Expected error not '
                                        'found: "{}"'.format(xerr))
        # Add remaining errors to item's error list
        if 'ERROR' in result and len(result['ERROR']) > 0:
            self.errors.extend(result['ERROR'])
        if len(self.errors) > 0:
            if self.config.getoption("--create-macros"):
                macro_file = 'run_{}.anymcr'.format(self.name)
                macro_file = os.path.join(self.fspath.dirname, macro_file)
                with open(macro_file,'w') as f:
                    f.writelines([str(mcr)+'\n' for mcr in self.macro])
                self.macro_file = macro_file
            raise  AnyException(self)
        return
        
    def repr_failure(self, excinfo):
        """ called when self.runtest() raises an exception. """
        if isinstance(excinfo.value, AnyException):
            rtn = 'Execution failed:\n'
            for elem in self.errors:
                rtn += textwrap.fill(elem, 80, 
                                     initial_indent='  *',
                                     subsequent_indent='   ')
                rtn += '\n'
            rtn += "\nMain file:\n"
            rtn += "  {}\n".format(self.fspath.strpath.replace(os.sep,os.altsep))
            rtn += "AnyBody Console:\n"
            rtn += "  {}\n".format(self.app.anybodycon_path.replace(os.sep, os.altsep))
            rtn += "Special model configuration:\n"
            for k,v in self.defs.items():
                rtn += "  #define {} {}\n".format(k,v)
            for k,v in self.paths.items():
                rtn += "  #path {} {}\n".format(k,v)
            if self.macro_file is not None:
                macro_file = self.macro_file.replace(os.sep,os.altsep)
                rtn += 'Macro:\n'
                rtn += '  anybody.exe -m "{}" &\n'.format(macro_file)
            return rtn
        else:
            return str(excinfo.value)

    def reportinfo(self):
        return self.fspath, 0, "AnyBody Simulation: %s" % self.name

        
class AnyException(Exception):
    """ custom exception for error reporting. """


    
def get_ammr_version(ammr_path):
    import xml.etree.ElementTree as ET
    version_file = os.path.join(ammr_path,'AMMR.version.xml')
    try:
        tree = ET.parse(version_file)
        version = tree.getroot()
        vstring = "{}.{}.{}".format(version.find('v1').text,
                                    version.find('v2').text,
                                    version.find('v3').text )
    except:
        vstring = "Unknown AMMR version"
    return vstring


def copy_files(src_dir, dst_dir):
    src_dir = str(src_dir)
    dst_dir = str(dst_dir)
    dirlist = os.listdir( src_dir )
    for name in dirlist:
        if name.endswith('.py') and name.find('test') != -1:
            continue
        full_name = os.path.join( src_dir, name)
        if (os.path.isfile(full_name) and name.find('_test')):
            shutil.copy(full_name, dst_dir)
        if os.path.isdir(full_name):
            dst_subdir = os.path.join(dst_dir, name)
            shutil.copytree(full_name,dst_subdir, 
                            ignore=shutil.ignore_patterns('test_*.py','*_test.py'))

        
@pytest.fixture(scope='module')
def copyfiles(request, test_dir):
    if request.config.getoption('--inplace') or request.config.getoption('--copyfiles'):
        # No need to copy files if test is allready run inplace, or if it
        # is done as a global option.
        pass
    else:
        model_folder = request.fspath.new(basename='')
        copy_files(model_folder, test_dir)
    return test_dir


@pytest.fixture(scope='module')
def test_dir(request):
    model_folder = request.fspath.new(basename='')
    if ( request.config.getoption('--inplace') 
         and not request.config.getoption('--copyfiles')):
        return model_folder
    else:
        tempdir = request.config._tmpdirhandler.mktemp(model_folder.basename,
                                                   numbered=True)
        return tempdir

        
@pytest.fixture(scope='module')
def model_dir(request, test_dir):
    model_folder = request.fspath.new(basename='')
    if (request.config.getoption('--copyfiles') ):
        copy_files(model_folder, test_dir)
        model_folder = test_dir
    return model_folder


def pytest_addoption(parser):
    group = parser.getgroup("anypytools", "testing AnyBody models")

    group._addoption("--ammr", action="store", default="built-in", metavar="path",
        help="AMMR used in test: built-in or path-to-ammr")
    group._addoption("--anybodycon", action="store", metavar="path",
        help="anybodycon.exe used in test: default or path-to-anybodycon")
    group._addoption("--collect-main-files", action="store_true",
        help="Also collect any files called ending in 'main.any'")
    group._addoption("--only-load", action="store_true",
        help="Only run a load test. I.e. do not run the 'RunTest' macro")
    group._addoption("--run-macro", action="store", default=None,
        help="Specify a special macro to run")
    group._addoption("--create-macros", action="store_true",
        help="Create a macro file if the test fails. This makes it easy to re-run "
             "the failed test in the gui application.")
    group._addoption("--inplace", action="store_true",
        help="Run tests in place. The macro file will be placed together with "
             "the model files. It becomes the responsobility of the model to "
             "ensure that the correct path statements are set.")
    group._addoption("--copyfiles", action="store_true",
        help=("Copy all test files to temp dir before test." 
              "This option has no effect if --inplace is set.") )
    group._addoption("--ammrdirs", action="append", default=[], metavar="path",
           help="add AMMR path to test for. (still under test)")
    
    parser.addini('ammrdirs', 'list of ammr paths to test against.', type="pathlist")


#def pytest_report_header(config):
#    return '\nAnyPyTools Test Plugin\n'
#    


def pytest_configure(config):
    anybodycon_path = config.getoption("--anybodycon")
    if anybodycon_path is None:
        anybodycon_path = get_anybodycon_path()

    if not os.path.isfile(anybodycon_path):
        raise IOError('Cound not find: {}'.format(anybodycon_path))
    
    console_output = subprocess.check_output([anybodycon_path, '/ni'])
    ams_version = console_output.splitlines()[2].decode()[23:]
    setattr(config, 'anybodycon_version', ams_version)
    setattr(config,'anybodycon_path', anybodycon_path)
    

class AnyTestFixture():
    def __init__(self, test_dir, model_dir, app, ammr):
        self.app = app
        self.ammr = ammr
        self.model_path = model_dir
        self.test_dir = test_dir
        self.macro_gen = MacroGenerator()
        self.path_kw = {'AMMR_PATH':ammr,
                        'TEMP_PATH': test_dir, 
                        'ANYBODY_PATH_OUTPUT':test_dir}
        self.define_kw = {}

       
    def load_macro(self,mainfile,define={},path={}):
        main_path = os.path.join(self.model_path,mainfile)
        load_str = 'load "{}" '.format(main_path)
        self.path_kw.update(path)
        self.define_kw.update(define)
        for key,value in self.path_kw.items():
            load_str += self.path2str(key,value)+" "
        for key,value in self.define_kw.items():
            load_str += self.define2str(key,value)+" "
        return load_str

    def check_output_log(self,result_list):
        __tracebackhide__ = True
        for result in result_list:
            if 'ERROR' in result:
                for err in result['ERROR']:
                    pytest.fail(err)
    
    def check_model_load_failure(self, result_list):
        __tracebackhide__ = True
        for result in result_list:
            if 'ERROR' in result:
                for err in result['ERROR']:
                    if 'Model loading skipped' in err:
                        pytest.fail(err)
    
    
    def define2str(self,key,value=None):
        if isinstance(value, str):
            if value.startswith('"') and value.endswith('"'):
                defstr = '-def %s=---"\\"%s\\""'% (key, value[1:-1])
            else:
                defstr = '-def %s="%s"'% (key, value)
        elif value is None:
            defstr = '-def %s=""'% (key)
        elif isinstance(value,float) :
            defstr =  '-def %s="%g"'% (key, value) 
        else:
            defstr = '-def %s="%d"'% (key, value) 
        return defstr 
        
    def path2str(self,key,path='.'):
        return '-p %s=---"%s"'% (key, path.replace('\\','\\\\')) 
        
        
        
        
@pytest.fixture(scope='session')
def ammr(request):
    ammr_path = request.config.getoption("--ammr")
    if ammr_path == 'built-in':
        ammr_path = os.path.join(os.path.dirname(get_anybodycon_path()),'AMMR')
    if not os.path.exists(ammr_path):
        raise IOError('Cound not find: {}'.format(ammr_path))
    ammr_version = get_ammr_version(ammr_path)

    return ammr_path
        
        
@pytest.fixture(scope='session')
def anybodycon(request):
    anybodycon_path = request.config.getoption("--anybodycon")
    if anybodycon_path == None:
        anybodycon_path = get_anybodycon_path()
    if not os.path.isfile(anybodycon_path):
        raise IOError('Cound not find: {}'.format(anybodycon_path))
    console_output = subprocess.check_output([anybodycon_path, '/ni'])
    ams_version = console_output.splitlines()[2].decode()[23:]
    return anybodycon_path
        

        
@pytest.yield_fixture()
def anytest(request, test_dir, model_dir, ammr, anybodycon):  
    #global _anybodycon_path, _ammr_path
    
    ammr_path = ammr
    abc_path = anybodycon
    
    if (request.config.getoption('--inplace') 
          and not request.config.getoption('--copyfiles') ):
        # Don't keep log files if test are not copied and inplace option is set.
        # This avoids clutter in the model directory
        keep_files = False
    else:
        keep_files = True
        
    
    
    app = AnyPyProcess( anybodycon_path = abc_path,
                        keep_logfiles = keep_files,
                        logfile_prefix = request.function.__name__, 
                        disp=False)
    
    atf = AnyTestFixture(str(test_dir), str(model_dir), app, ammr_path)
    if request.config.getoption('--inplace'):
        atf.path_kw = {}
        atf.model_path = ''
        
    
    with test_dir.as_cwd():
        yield atf
